"""Tests for model trace forest structure when cavities which overlap are included."""

from itertools import permutations

import pytest
from finesse import Model
from finesse.tracing.tree import TraceTree
from testutils.diff import assert_trace_trees_equal


@pytest.fixture()
def aligo_model():
    IFO = Model()
    IFO.parse(
        """
    # modulators for core interferometer sensing - Advanced LIGO, CQG, 2015
    # http://iopscience.iop.org/article/10.1088/0264-9381/32/7/074001/meta#cqg507871s4-8
    # 9MHz (CARM, PRC, SRC loops)
    variable f1 9099471
    variable f2 5*f1
    variable nsilica 1.45
    variable Mloss 30u

    ###############################################################################
    ###   length definitions
    ###############################################################################
    variable Larm 3994
    variable LPR23 16.164  # distance between PR2 and PR3
    variable LSR23 15.443  # distance between SR2 and SR3
    variable LPR3BS 19.538 # distance between PR3 and BS
    variable LSR3BS 19.366 # distance between SR3 and BS
    variable lmich 5.342   # average length of MICH
    variable lschnupp 0.08
    variable lPRC (3+0.5)*c0/(2*f1) # T1000298 Eq2.1, N=3
    variable lSRC (17)*c0/(2*f2) # T1000298 Eq2.2, M=3

    ###############################################################################
    ###   laser
    ###############################################################################
    laser L0 P=125
    mod mod1 f=f1 midx=0.18 order=1 mod_type=pm
    mod mod2 f=f2 midx=0.18 order=1 mod_type=pm
    link(L0, mod1, mod2)

    ###############################################################################
    ###   PRC
    ###############################################################################
    s sPRCin mod2.p2 PRMAR.p1
    m PRMAR R=0 L=40u xbeta=PRM.xbeta ybeta=PRM.ybeta phi=PRM.phi
    s sPRMsub1 PRMAR.p2 PRM.p1 L=0.0737 nr=nsilica
    m PRM T=0.03 L=8.5u Rc=11.009
    s lp1 PRM.p2 PR2.p1 L=lPRC-LPR3BS-LPR23-lmich
    bs PR2 T=250u L=Mloss alpha=-0.79 Rc=-4.545
    s lp2 PR2.p2 PR3.p1 L=LPR23
    bs PR3 T=0 L=Mloss alpha=0.615 Rc=36.027
    s lp3 PR3.p2 BS.p1 L=LPR3BS

    ###############################################################################
    ###   BS
    ###############################################################################
    bs BS R=0.5 L=Mloss alpha=45
    s BSsub1 BS.p3 BSAR1.p1 L=0.0687 nr=nsilica
    s BSsub2 BS.p4 BSAR2.p2 L=0.0687 nr=nsilica
    bs BSAR1 L=50u R=0 alpha=-29.195
    bs BSAR2 L=50u R=0 alpha=29.195

    ###############################################################################
    ###   Yarm
    ###############################################################################
    # Distance from beam splitter to Y arm input mirror
    s ly1 BS.p2 ITMYlens.p1 L=lmich-lschnupp/2-ITMYsub.L*ITMXsub.nr
    lens ITMYlens f=34500
    s ly2 ITMYlens.p2 ITMYAR.p1
    m ITMYAR R=0 L=20u xbeta=ITMY.xbeta ybeta=ITMY.ybeta phi=ITMY.phi
    s ITMYsub ITMYAR.p2 ITMY.p1 L=0.2 nr=nsilica
    m ITMY T=0.014 L=Mloss Rc=-1934
    s LY ITMY.p2 ETMY.p1 L=Larm
    m ETMY T=5u L=Mloss Rc=2245
    s ETMYsub ETMY.p2 ETMYAR.p1 L=0.2 nr=nsilica
    m ETMYAR 0 500u xbeta=ETMY.xbeta ybeta=ETMY.ybeta phi=ETMY.phi
    cav cavYARM ETMY.p1.o

    ###############################################################################
    ###   Xarm
    ###############################################################################
    # Distance from beam splitter to X arm input mirror
    s lx1 BSAR1.p3 ITMXlens.p1 L=lmich+lschnupp/2-ITMXsub.L*ITMXsub.nr-BSsub1.L*BSsub1.nr
    lens ITMXlens f=34500
    s lx2 ITMXlens.p2 ITMXAR.p1
    m ITMXAR R=0 L=20u xbeta=ITMX.xbeta ybeta=ITMX.ybeta phi=ITMX.phi
    s ITMXsub ITMXAR.p2 ITMX.p1 L=0.2 nr=nsilica
    m ITMX T=0.014 L=Mloss Rc=-1934
    s LX ITMX.p2 ETMX.p1 L=Larm
    m ETMX T=5u L=Mloss Rc=2245
    s ETMXsub ETMX.p2 ETMXAR.p1 L=0.2 nr=nsilica
    m ETMXAR 0 500u xbeta=ETMX.xbeta ybeta=ETMX.ybeta phi=ETMX.phi
    cav cavXARM ETMX.p1.o

    ###############################################################################
    ###   SRC
    ###############################################################################
    s ls3 BSAR2.p4 SR3.p1 L=LSR3BS
    bs SR3 T=0 L=Mloss alpha=0.785 Rc=35.972841
    s ls2 SR3.p2 SR2.p1 L=LSR23
    bs SR2 T=0 L=Mloss alpha=-0.87 Rc=-6.406
    s ls1 SR2.p2 SRM.p1 L=lSRC-LSR3BS-LSR23-BSsub2.L*BSsub2.nr-lmich
    m SRM T=0.2 L=8.7u Rc=-5.6938
    s SRMsub SRM.p2 SRMAR.p1 L=0.0749 nr=nsilica
    m SRMAR R=0 L=50n

    ###############################################################################
    ###   Output path
    ###############################################################################
    # Here we just use some simple filter to approximate an OMC for filtering
    # out RF fields, this doesn't filter HOMs!
    dbs OFI
    sq sqz db=6 angle=90.0
    link(sqz, OFI.p2)

    # (as built parameters: D1300507-v1)
    s sSRM_OFI SRMAR.p2 OFI.p1 L=0.7278
    s sOFI_OM1 OFI.p3 OM1.p1 L=2.9339

    bs OM1 T=800u L=0 alpha=2.251 Rc=[4.6, 4.6]
    s sOM1_OM2 OM1.p2 OM2.p1 L=1.395
    bs OM2 T=0 L=0 alpha=4.399 Rc=[1.7058, 1.7058]
    s sOM2_OM3 OM2.p2 OM3.p1 L=0.631
    bs OM3 T=0 L=0 alpha=30.037
    s sOM3_OMC OM3.p2 OMC_IC.p1 L=0.2034

    ###############################################################################
    ###   OMC
    ###############################################################################
    # obp OMC fc=0 bandwidth=1M filter_hom=[0,0]
    # link(SRMAR.p2, OFI.p1)
    # link(OFI.p3, OMC)

    cav cavOMC OMC_IC.p3.o
    bs OMC_IC T=0.0076 L=10u alpha=2.7609
    s lIC_OC OMC_IC.p3 OMC_OC.p1 L=0.2815
    bs OMC_OC T=0.0075 L=10u alpha=4.004
    s lOC_CM1 OMC_OC.p2 OMC_CM1.p1 L=0.2842
    bs OMC_CM1 T=36u L=10u alpha=4.004 Rc=[2.57321, 2.57321]
    s lCM1_CM2 OMC_CM1.p2 OMC_CM2.p1 L=0.2815
    bs OMC_CM2 T=35.9u L=10u alpha=4.004 Rc=[2.57369, 2.57369]
    s lCM2_IC OMC_CM2.p2 OMC_IC.p4 L=0.2842

    ###############################################################################
    ### Length sensing and control
    ###############################################################################
    dof XARM ETMX.dofs.z
    dof YARM ETMY.dofs.z
    dof CARM ETMX.dofs.z +1 ETMY.dofs.z +1
    dof DARM ETMX.dofs.z +1 ETMY.dofs.z -1
    dof PRCL PRM.dofs.z +1
    dof SRCL SRM.dofs.z +1 DC=90
    dof MICH BS.dofs.z +1
    dof MICH2 ITMY.dofs.z +1 ETMY.dofs.z +1 ITMX.dofs.z -1 ETMX.dofs.z -1
    dof STRAIN LX.dofs.h +1 LY.dofs.h -1
    dof FRQ L0.dofs.frq
    dof RIN L0.dofs.amp

    readout_rf REFL9 PRMAR.p1.o f=f1
    readout_rf REFL18 PRMAR.p1.o f=3*f1
    readout_rf REFL45 PRMAR.p1.o f=5*f1
    readout_rf POP9  PR2.p3.o   f=f1
    readout_rf POP45 PR2.p3.o   f=f2
    readout_rf AS45  SRMAR.p2.o f=f2
    readout_dc AS    OMC_OC.p3.o

    lock CARM_lock REFL9.outputs.I CARM.DC -0.1 1e-6
    lock MICH_lock POP45.outputs.Q MICH.DC -15 1e-6
    lock PRCL_lock POP9.outputs.I PRCL.DC 2.8 1e-6
    lock SRCL_lock POP45.outputs.I SRCL.DC 42 1e-6
    lock DARM_rf_lock AS45.outputs.I DARM.DC -0.003 1e-6
    lock DARM_dc_lock AS.outputs.DC DARM.DC -0.003 1e-6 offset=20m enabled=False

    ###############################################################################
    ### DC power measurements
    ###############################################################################
    pd Px ETMX.p1.i
    pd Py ETMX.p1.i
    pd Pprc PRM.p2.o
    pd Psrc SRM.p1.i
    pd Prefl ETMX.p1.i
    pd Pas OMC_OC.p3.o
    """
    )
    return IFO


@pytest.mark.parametrize("order", list(permutations(["PRX", "PRY"])))
def test_overlapping_prx_pry_forest(aligo_model: Model, order):
    """Test that overlapping power recycling cavities in X and Y arms are structured
    correctly in the model trace forest."""
    IFO = aligo_model

    IFO.parse(
        """
    cav PRX PRM.p2 ITMX.p1.i
    cav PRY PRM.p2 ITMY.p1.i
    """
    )
    IFO.remove("cavOMC")
    IFO.beam_trace(order=order)

    expect = [TraceTree.from_cavity(getattr(IFO, cav)) for cav in order]
    expect.reverse()

    # Arm cavities don't overlap with anything so will constitute
    # the first block of internal trace trees
    expect.insert(0, TraceTree.from_cavity(IFO.cavXARM))
    expect.insert(1, TraceTree.from_cavity(IFO.cavYARM))

    # Only really care about the internal trace trees, other
    # tests cover external forest structures
    for i in range(len(expect)):
        assert_trace_trees_equal(IFO.trace_forest.forest[i], expect[i])


@pytest.mark.parametrize("order", list(permutations(["SRX", "SRY"])))
def test_overlapping_srx_sry_forest(aligo_model: Model, order):
    """Test that overlapping signal recycling cavities in X and Y arms are structured
    correctly in the model trace forest."""
    IFO = aligo_model

    IFO.remove("cavOMC")
    IFO.parse(
        """
    cav SRX SRM.p1 ITMX.p1.i
    cav SRY SRM.p1 ITMY.p1.i
    """
    )

    IFO.beam_trace(order=order)

    expect = [TraceTree.from_cavity(getattr(IFO, cav)) for cav in order]
    expect.reverse()

    # Arm cavities don't overlap with anything so will constitute
    # the first block of internal trace trees
    expect.insert(0, TraceTree.from_cavity(IFO.cavXARM))
    expect.insert(1, TraceTree.from_cavity(IFO.cavYARM))

    # Only really care about the internal trace trees, other
    # tests cover external forest structures
    for i in range(len(expect)):
        assert_trace_trees_equal(IFO.trace_forest.forest[i], expect[i])


@pytest.mark.parametrize("order", list(permutations(["SRX", "SRY", "PRX", "PRY"])))
def test_overlapping_all_recycling_cavs_forest(aligo_model: Model, order):
    """Test that overlapping both recycling cavities in X and Y arms are structured
    correctly in the model trace forest."""
    IFO = aligo_model

    IFO.parse(
        """
    cav SRX SRM.p1 ITMX.p1.i
    cav SRY SRM.p1 ITMY.p1.i
    cav PRX PRM.p2 ITMX.p1.i
    cav PRY PRM.p2 ITMY.p1.i
    """
    )
    IFO.remove("cavOMC")
    IFO.remove("cavXARM")
    IFO.remove("cavYARM")

    IFO.beam_trace(order=order)

    expect = [TraceTree.from_cavity(getattr(IFO, cav)) for cav in order]
    expect.reverse()

    # Only really care about the internal trace trees, other
    # tests cover external forest structures
    for i in range(len(expect)):
        assert_trace_trees_equal(IFO.trace_forest.forest[i], expect[i])



"""Higher order functions."""

from functools import update_wrapper
from enum import Enum


class valuedispatchmethod:
    """Method decorator that performs single-dispatch based on value.

    Transforms a method into a generic method, which can have different behaviours
    depending upon the value of its first argument (after `self`). The decorated
    method acts as the default implementation, and additional implementations can be
    registered using the register() attribute of the generic method.

    The default implementation must accept at least one argument, the value. Specialised
    implementations are not passed the value. Both implementations are passed any
    additional positional and keyword arguments specified in the call to `get`.

    Values registered via register() must be hashable.

    Based on :class:`functools.singledispatchmethod`.

    Examples
    --------
    Define a class with a value dispatch method, register a method to handle particular
    values, and call it:

    >>> from finesse.utilities.functools import valuedispatchmethod
    >>> class Example:
    ...     @valuedispatchmethod
    ...     def get(self, value):
    ...          return f"got {value} (default)"
    ...     @get.register(1)
    ...     def _(self):
    ...         return "got 1"
    ...
    >>> myobj = Example()
    >>> myobj.get(1)
    got 1
    >>> myobj.get(2)
    got 2 (default)
    """

    def __init__(self, func):
        if not callable(func) and not hasattr(func, "__get__"):
            raise TypeError(f"{func!r} is not callable or a descriptor")

        self.func = func
        self.registry = {}

    def dispatch(self, value):
        return self.registry.get(value)

    def _validate_key(self, key):
        pass

    def register(self, value):
        """Register a new implementation of the generic method for the given value."""
        self._validate_key(value)

        def wrap(method):
            self.registry[value] = method
            return method

        return wrap

    def __get__(self, obj, cls=None):
        def _method(*args, **kwargs):
            args = list(args)
            value = args.pop(0)

            if (method := self.dispatch(value)) is not None:
                return method.__get__(obj, cls)(*args, **kwargs)
            else:
                # Call the default with the value.
                return self.func(cls, value, *args, **kwargs)

        _method.__isabstractmethod__ = self.__isabstractmethod__
        _method.register = self.register
        update_wrapper(_method, self.func)
        return _method

    @property
    def __isabstractmethod__(self):
        return getattr(self.func, "__isabstractmethod__", False)


class flagdispatchmethod(valuedispatchmethod):
    """Method decorator that performs single-dispatch based on a flag enumeration.

    Transforms a method into a generic method, which can have different behaviours
    depending upon the value of its first argument (after `self`). The decorated
    method acts as the default implementation, and additional implementations can be
    registered using the register() attribute of the generic method.

    The default implementation must accept at least one argument, the flags. Specialised
    implementations are not passed the flags. Both implementations are passed any
    additional positional and keyword arguments specified in the call to `get`.

    As long as the given flag is contained within a registered flag enumeration, it
    triggers that corresponding method. The method corresponding to the first registered
    flag that matches is returned.

    Based on :class:`functools.singledispatchmethod`.

    Examples
    --------
    Define a class with a flag dispatch method, register a method to handle particular
    flags, and call it:

    >>> from enum import auto, Flag
    >>> from finesse.utilities.functools import flagdispatchmethod
    >>> class Flags(Flag):
    ...     A = auto()
    ...     B = auto()
    ...     C = auto()
    ...     D = auto()
    ...
    >>> class Example:
    ...     @flagdispatchmethod
    ...     def get(self, flag):
    ...          return f"got {flag} (default)"
    ...     @get.register(Flags.A)
    ...     def _(self):
    ...         return f"got {Flags.A}"
    ...     @get.register(Flags.B)
    ...     @get.register(Flags.C)
    ...     def _(self):
    ...         return f"got {Flags.B}"
    ...
    >>> myobj = Example()
    >>> myobj.get(Flags.A)
    got Flags.A
    >>> myobj.get(Flags.A | Flags.B)
    got Flags.B|A (default)
    >>> myobj.get(Flags.C)
    got Flags.B
    >>> myobj.get(Flags.D)
    got Flags.D (default)
    """

    def _validate_key(self, key):
        if not isinstance(key, Enum):
            raise ValueError(f"{self.__class__.__name__} requires Enum values")

    def dispatch(self, enumeration):
        for key, method in self.registry.items():
            if enumeration in key:
                return method



"""Random collection of Actions that do no warrant a separate module."""

from ...parameter import Parameter
from .base import Action, convert_str_to_parameter
from finesse.solutions import BaseSolution

import logging

LOGGER = logging.getLogger(__name__)


class SaveModelAttrSolution(BaseSolution):
    """
    Attributes
    ----------
    values : dict
        Dictionary of model attribute values
    """


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Plot(Action):
    def __init__(self, name="abcd"):
        super().__init__(name)

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        raise NotImplementedError()


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Printer(Action):
    def __init__(self, *args, name="printer", eval=True):
        super().__init__(name)
        self.args = args
        self._eval = eval

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        if self._eval:
            out = []
            for _ in self.args:
                if hasattr(_, "eval"):
                    out.append(_.eval())
                else:
                    out.append(_)
            print(*out)
        else:
            print(*(_ for _ in self.args))


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class PrintModel(Action):
    """An action that prints the model object being currently used to run actions."""

    def __init__(self, name="print_model"):
        super().__init__(name)

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        print(state.model)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class StoreModelAttr(Action):
    def __init__(self, *args):
        super().__init__(self.__class__.__name__)
        self.args = tuple(a if isinstance(a, str) else a.full_name for a in args)

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        sol = SaveModelAttrSolution(self.name)
        sol.values = {}
        for _ in self.args:
            p = state.model.get(_)
            if hasattr(p, "eval"):
                sol.values[_] = p.eval()
            else:
                sol.values[_] = p
        return sol


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class PrintModelAttr(Action):
    """Prints an attribute of the model being currently used.

    Parameters
    ----------
    \\*args : (str,)
        Strings input for the attribute to print

    eval : bool, optional
        When `True` symbolic expressions will be evaluated before printing.
        Defaults to `True`.

    prefix : str, optional
        Optional string to print before the attributes

    Examples
    --------
    You can print the current value of parameters and such using:

    >>> PrintModelAttr("m1.R", "bs.phi")
    """

    def __init__(self, *args, eval=True, prefix=""):
        super().__init__(self.__class__.__name__)
        self.args = tuple(a if isinstance(a, str) else a.full_name for a in args)
        self.prefix = prefix
        self._eval = eval

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        out = [self.prefix]
        for _ in self.args:
            obj = state.model.get(_)
            if hasattr(obj, "eval") and self._eval:
                out.append(f"{_}={obj.eval()}")
            else:
                out.append(f"{_}={obj}")
        print(*out)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Change(Action):
    """Changes a model Parameter to some value during an analysis.

    Parameters
    ----------
    change_dict : dict, optional
        Dictionary of parameter:value pairs to change.
    relative : bool, optional
        Whether to increment from the parameters current value or not
    name : str, optional
        Name of action
    \\**kwargs
        Alternative method to specify parameter:value pairs to change

    Examples
    --------
    A simple change of a parameter between running two `noxaxis` analyses:

    >>> model = finesse.script.parse("l L1 P=1")
    >>> model.run('series(noxaxis(), change(L1.P=2), noxaxis())')

    Or increment from the current value:
    >>> model.run('series(noxaxis(), change(L1.P=1, relative=True), noxaxis())')
    """

    def __init__(self, change_dict=None, *, relative=False, name="change", **kwargs):
        super().__init__(name)
        self.change_dict = change_dict
        self.kwargs = kwargs
        self.relative = relative

    @property
    def change_kwargs(self):
        kwargs = self.kwargs or {}
        if self.change_dict:
            kwargs.update(self.change_dict)
        return kwargs

    def _requests(self, model, memo, first=True):
        for el in self.change_kwargs.keys():
            p = convert_str_to_parameter(model, el)
            if isinstance(p, Parameter):
                memo["changing_parameters"].append(el)
            else:
                raise TypeError(
                    f"{el} is not a name of a Parameter or Component in the model"
                )

    def _do(self, state):
        for el, val in self.change_kwargs.items():
            p = convert_str_to_parameter(state.model, el)
            if self.relative:
                p.value += val
            else:
                p.value = val


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Execute(Action):
    """An action that will execute the function passed to it when it is run.

    Parameters
    ----------
    do_fn : function
        A function that takes an AnalysisState, and the name of the Exec action as its
        only arguments. If this function returns a
        :class:`finesse.solutions.base.BaseSolution` object then it will be added to the
        simulations solution to return to the user.
    parameters : list, optional
        A list of parameters that will be changed by do_fn, if any.
    name : str
        The name to give this action.

    Examples
    --------
    A simple function to execute might use a pattern such as this, which generates
    a Solution that is returned back to the user.

    >>> from finesse.solutions import SimpleSolution
    >>> def my_action(state, name):
    ...     sol = SimpleSolution(name)
    ...     return sol

    The :class:`finesse.solutions.simple.SimpleSolution` object is just an object you
    can store anything you want in. You can extract any state information about the
    simulation or model and store it here. This allows you to probe and store details
    that might not be available as a detector, for example.
    """

    def __init__(self, do_fn, parameters=None, name="execute"):
        super().__init__(name)
        self.do_fn = do_fn
        self.parameters = parameters

    def _do(self, state):
        sol = self.do_fn(state, self.name)
        if isinstance(sol, BaseSolution):
            return sol

    def _requests(self, model, memo, first=True):
        if self.parameters is not None:
            memo["changing_parameters"].extend(self.parameters)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class UpdateMaps(Action):
    """Update any maps that might be changing in the simulation."""

    def __init__(self, name="update_maps", *args, **kwargs):
        super().__init__(name)
        self.args = args
        self.kwargs = kwargs

    def _requests(self, model, memo, first=True):
        return None

    def _do(self, state):
        state.sim.update_map_data()


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class LogModelAttribute(Action):
    def __init__(self, *attrs):
        super().__init__("print_parmeter")
        self.attrs = attrs

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        LOGGER.info(*(f"{_}={state.model.get(str(_))}" for _ in self.attrs))


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Scale(Action):
    """Action for scaling simulation outputs by some fixed amount. Included for
    compatibility with legacy Finesse code. New users should apply any desired scalings
    manually from Python.

    Parameters
    ----------
    detectors : dict
        A dictionary of `detector name: scaling factor` mappings.
    """

    def __init__(self, scales: dict, **kwargs):
        super().__init__(None)
        self.kwargs = kwargs
        self.scales = scales

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        sol = state.previous_solution
        for det, fac in self.scales.items():
            sol._outputs[det][()] *= fac


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class MakeTransparent(Action):
    """Action to make all provided surfaces transparent. Simply sets the reflectivity to
    zero and transmitivity to one.

    Parameters
    ----------
    surfaces : list
        A list of surface component names to be made transparent.
    """

    def __init__(self, surfaces, name="make transparent"):
        super().__init__(name)
        self.surfaces = surfaces

    def _do(self, state):
        for name, el in state.sim.model.elements.items():
            if name in self.surfaces:
                el.set_RTL(R=0, T=1)

    def _requests(self, model, memo, first=True):
        for name, el in model.elements.items():
            if name in self.surfaces:
                memo["changing_parameters"].extend(
                    [el.R.full_name, el.T.full_name, el.L.full_name]
                )



"""Miscellaneous utility functions for any part of Finesse."""

from os import fspath, PathLike
import functools
import re
from itertools import tee
from contextlib import contextmanager, closing, nullcontext
from functools import partial, reduce

import numpy as np


def reduce_getattr(obj, key: str, delimiter: str = "."):
    """Applies a nested getattr with reduce to select an attribute of a nested object
    within `obj`.

    Parameters
    ----------
    obj : object
        Object to search

    key : str
        Delimited string of attributes

    delimiter : str, optional
        Delimiter character of key

    Returns
    -------
    Attribute of object
    """
    attrs = key.strip().split(delimiter)
    return reduce(getattr, attrs, obj)


def calltracker(func):
    """Decorator used for keeping track of whether the current state is inside the
    decorated function or not.

    Sets an attribute `has_been_called` on the function which gets switched on when the
    function is being executed and switched off after the function has returned. This
    allows you to query ``func.has_been_called`` for determining whether the code being
    executed has been called from within `func`.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        wrapper.has_been_called = True
        out = func(*args, **kwargs)
        wrapper.has_been_called = False

        return out

    wrapper.has_been_called = False
    return wrapper


def valid_name(name):
    """Validate the specified name."""
    return re.compile("^[a-zA-Z_][a-zA-Z0-9_]*$").match(name)


def check_name(name):
    """Checks the validity of a component or node name.

    A name is valid if it contains only alphanumeric characters and underscores, and is
    not empty.

    Parameters
    ----------
    name : str
        The name to check.

    Returns
    -------
    name : str
        The name passed to this function if valid.

    Raises
    ------
    ValueError
        If `name` contains non-alphanumeric / underscore characters.
    """
    if not valid_name(name):
        raise ValueError(
            f"'{name}' can only contain alphanumeric and underscore characters"
        )
    return name


def pairwise(iterable):
    """Iterates through each pair in a iterable.

    Parameters
    ----------
    iterable : :py:class:`collections.abc.Iterable`
        An iterable object.

    Returns
    -------
    zip
        A zip object whose `.next()` method returns a tuple where the i-th
        element comes from the i-th iterable argument.
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def find(x, value):
    """Finds `value` in the list `x` and returns its index, returning `None` if `value`
    is not in the list."""
    try:
        return x.index(value)
    except ValueError:
        return None


def find_nearest(x, value, index=False):
    idx = np.argmin(np.abs(x - value))
    if index:
        return idx
    return x[idx]


def is_iterable(obj):
    """Reliable check for whether an object is iterable.

    Note that strings are treated as non-iterable objects
    when performing this check. This will only return true
    for iterable non-str objects.

    Returns
    -------
    flag : bool
        True if `obj` is iterable, False otherwise.
    """
    try:
        iter(obj)
    except Exception:
        return False
    else:
        return not isinstance(obj, str)


@contextmanager
def opened_file(filename, mode):
    """Get an open file regardless of whether a string or an already open file is
    passed.

    Adapted from :func:`numpy.loadtxt`.

    Parameters
    ----------
    filename : str, :class:`pathlib.Path`, or file-like
        The path or file object to ensure is open. If `filename` is an already open file
        object, it is yielded as-is, and is *not* closed after the wrapped context
        exits. If `filename` is a string, it is opened with the specified `mode` and
        yielded, then closed once the wrapped context exits.

    mode : str
        The mode to open `filename` with, if it is not already open.

    Yields
    ------
    :class:`io.FileIO`
        The open file with the specified `mode`.

    Notes
    -----
    If `filename` is an open file, `mode` is ignored; it is the responsibility of the
    calling code to check that it is opened with the correct mode.
    """
    if isinstance(filename, PathLike):
        filename = fspath(filename)

    if isinstance(filename, str):
        fid = open(filename, mode)
        fid_context = closing(fid)
    else:
        fid = filename
        fid_context = nullcontext(fid)

    with fid_context:
        yield fid


def graph_layouts():
    """Available NetworkX and graphviz (if installed) graph plotting layouts."""
    return {**networkx_layouts(), **graphviz_layouts()}


def networkx_layouts():
    """Available NetworkX graph plotting layouts."""
    import inspect
    import networkx

    # Excluded layouts.
    excluded = (
        "rescale",
        # These layouts need the network to first be grouped into sets.
        "bipartite",
        "multipartite_layout",
    )

    layouts = {}
    suffix = "_layout"

    def find_layouts(members):
        for name, func in members:
            if name.startswith("_") or not name.endswith(suffix):
                # Not a public layout.
                continue

            # Strip the "_layout" part.
            name = name[: -len(suffix)]

            if name in excluded:
                continue

            layouts[name] = func

    find_layouts(inspect.getmembers(networkx.drawing.layout, inspect.isfunction))

    return layouts


def graphviz_layouts():
    """Available graphviz graph plotting layouts."""
    import networkx
    from ..env import has_pygraphviz

    layouts = {}

    if has_pygraphviz():
        for layout in ["neato", "dot", "fdp", "sfdp", "circo"]:
            # Returns callable that can be called like `networkx.drawing.layout` members.
            layouts[layout] = partial(
                networkx.drawing.nx_agraph.pygraphviz_layout, prog=layout
            )

    return layouts


def doc_element_parameter_table(cls):
    """Prints table for a particular element class."""
    import finesse
    from finesse.utilities.tables import Table
    from IPython.core.display import HTML

    def process_changeable(pinfo):
        if pinfo.changeable_during_simulation:
            return "✓"
        else:
            return "✗"

    headers = (
        "Name",
        "Description",
        "Units",
        "Data type",
        "Can change during simualation",
    )

    tbl = [headers] + [
        (p.name, p.description, p.units, p.dtype.__name__, process_changeable(p))
        for p in finesse.element.ModelElement._param_dict[cls][::-1]
    ]
    if len(tbl) == 0:
        raise RuntimeError(f"{cls} has no model parameters to display.")

    a = Table(
        tbl,
        headerrow=True,
        headercolumn=False,
        alignment=(("left", "left", "center", "center", "center"),),
    )
    for index, p in enumerate(finesse.element.ModelElement._param_dict[cls][::-1]):
        if p.changeable_during_simulation:
            a.color[index + 1, 4] = (0, 255, 0)
        else:
            a.color[index + 1, 4] = (255, 0, 0)
    return HTML(a.generate_html_table())


class DeprecationHelper:
    """Used for deprecating classes."""

    def __init__(self, old_name, new_name, new_target, until_version):
        self.new_target = new_target
        self.old_name = old_name
        self.new_name = new_name
        self.until_version = until_version

        self.msg = (
            f"{self.old_name} is deprecated as of version {self.until_version}: use "
            f"instead {self.new_name}"
        )

    def _warn(self):
        deprecation_warning(self.msg, self.until_version)

    def __call__(self, *args, **kwargs):
        self._warn()
        return self.new_target(*args, **kwargs)


def deprecation_warning(msg, until_version):
    """Function that warns a user about a deprecation. If the current version is past
    when this feature is deprecated it will raise `DeprecationWarning` instead.

    Parameters
    ----------
    msg : str
        Message to warn user with.
    until_version : str
        PEP 440 version string. After this version an exception is thrown.

    Raises
    ------
    DeprecationWarning
        When current version is >= `until_version`
    """
    from warnings import warn
    from packaging import version
    from finesse import __version__ as finesse_version

    if version.parse(finesse_version) >= version.parse(until_version):
        raise DeprecationWarning(msg)
    else:
        warn(msg, DeprecationWarning, stacklevel=4)


def inheritors(klass: type) -> set[type]:
    """Returns all classes that inherit from ``klass``

    Parameters
    ----------
    klass : type
        Class to get inheritors from

    Returns
    -------
    set[type]
        set of classes that are a subclass of ``klass``
    """
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses



cimport numpy as np
import numpy as np
cimport cython

from finesse.cymath.homs cimport in_mask
from finesse.detectors.general import Detector
from copy import copy

cdef class OutputInformation:
    def __init__(self, name, detector_type, nodes, dtype, unit, shape, label, needs_fields, needs_trace):
        if shape is None:
            shape = ()
        self.name = name
        self.__detector_type = detector_type
        self.__nodes = nodes
        self.__dtype = dtype
        self.__unit = unit
        self.__label = label
        self.__needs_fields = needs_fields
        self.__needs_trace = needs_trace
        self._update_dtype_shape(shape)

    def _set_dtype(self, dtype):
        self.__dtype = dtype

    @property
    def detector_type(self):
        """The type of the detector that has generated this output information."""
        return self.__detector_type

    @property
    def needs_fields(self):
        """Flag indicating whether the detector requires light fields (i.e. solving of
        the interferometer matrix)."""
        return self.__needs_fields

    @property
    def needs_trace(self):
        """Flag indicating whether the detector requires beam traces."""
        return self.__needs_trace

    @property
    def nodes(self):
        """The nodes this detector observes.

        :`getter`: Returns copy of detected nodes.
        """
        return copy(self.__nodes)

    @property
    def dtype(self):
        return self.__dtype

    @property
    def dtype_shape(self):
        return self.__dtype_shape

    def _update_dtype_shape(self, shape):
        """Only to be used internally by detectors like Cameras for updating the shape
        if resolution is changed. Shouldn't be changed by users directly or during a simulation."""
        self.__dtype_shape = shape
        self.__dtype_size = int(np.prod(shape))

    @property
    def dtype_size(self):
        """Size of the output in terms of number of elements.

        This is typically unity as most detectors return a single
        value via their output functions.

        Equivalent to the product of :attr:`.Detector.dtype_shape`.
        """
        return self.__dtype_size

    @property
    def unit(self):
        return self.__unit

    def _update_unit(self, unit):
        self.__unit = unit

    @property
    def label(self):
        return self.__label

    def _update_label(self, label):
        self.__label = label


cdef class OutputFuncWrapper:
    """Helper class for wrapping a C fill function that can be referenced from Python by
    objects. This allows a direct C call to the function from other cdef functions.

    Examples
    --------
    Create a C function then wrap it using this class:

    >>> cdef void c_output(DetectorWorkspace ptr_ws) noexcept:
    >>>    cdef PDWorkspace ws = <PDWorkspace>ptr_ws
    >>>    ...
    >>>
    >>> fill = OutputFuncWrapper.make_from_ptr(c_fill)
    """
    def __cinit__(self):
       self.func = NULL

    def __call__(self, ws):
        return self.func(ws)

    @staticmethod
    cdef OutputFuncWrapper make_from_ptr(fptr_c_output f) :
        cdef OutputFuncWrapper out = OutputFuncWrapper()
        out.func = f
        return out


cdef class DetectorWorkspace(ElementWorkspace):
    """A base class that all detector workspaces should inherit from. Provides a generic
    set of data needed to compute values and output them with metadata needed for
    storing the outputs.

    Parameters
    ----------
    owner : :class:`finesse.element.ModelElement`
        Detector `Element` that owns this workspace and will be setting it up
    sim : Simulation object
        Simulation object this workspace should be associated with
    values :  [object, :class:`finesse.element_workspace.BaseCValues`], optional
        The object containing the values that will be used by this workspace to
        calculate some output. These should match the parameters offered by the
        owner. A pure Python object can be used but will be slower to access.
        A :class:`finesse.element_workspace.BaseCValues` object can also be used that offers cythonised
        access to parameter values
    oinfo : OutputInformation, optional
        When provided this will set the output information of this detector, such
        as units, datatype, shape/dimension of outputs.
    needs_carrier : bool, optional
        If the carrier simulation data is needed, this must be True
    needs_signal : bool, optional
        If the signal simulation data (transfer functions) is needed, this must be True
    needs_noise : bool, optional
        If this detector requires noise covariances to be calculated this must be True
    needs_modal_update: bool, optional
        If this detector outputs some modal or geometric property, this must be True.

    Notes
    -----
    The `needs_*` flags specify which simulations should be run to evaluate this
    workspace. At least one should be True, unless `needs_simulation` is
    flagged as False. This is to catch certain cases which mean the workspace
    will just not produce any output.

    When adding new `needs_*` flag, ensure you update the `MathDetector` object
    to correctly fill these flags. The `MathDetector` essentially borrows workspaces
    from other detectors to compute its output and
    """

    def __init__(
            self,
            object owner,
            object sim,
            object values=None,
            *,
            OutputInformation oinfo=None,
            bint needs_carrier=False,
            bint needs_signal=False,
            bint needs_noise=False,
            bint needs_modal_update=False,
            bint needs_simulation=True):

        flags = (needs_noise or needs_signal or needs_carrier or needs_modal_update)
        if not flags and needs_simulation:
            raise Exception(f"Detector workspace {self} has the flags set so that `needs_noise or needs_signal or needs_carrier or needs_modal_update == False`, so no calculations will be done.")
        elif not needs_simulation and flags:
            raise Exception(f"Detector workspace {self} is marked as not needing a simulation result but is flagging that it does.")

        super().__init__(sim, owner, values)
        self.needs_noise = needs_noise
        self.needs_signal = needs_signal
        self.needs_carrier = needs_carrier
        self.needs_modal_update = needs_modal_update
        if oinfo is None and isinstance(owner, Detector):
            oinfo = owner.output_information
        elif oinfo is None:
            raise ValueError("output information object must be provided if owner is not a detector")

        self.oinfo = oinfo
        self.ignore_sim_mask = False

    cpdef get_output(self) :
        self.update_parameter_values()
        if self.fn_c is not None:
            return self.fn_c.func(self)
        elif self.fn_py is not None:
            return self.fn_py(self)
        else:
            return None

    def set_output_fn(self, callback):
        if type(callback) is OutputFuncWrapper:
            self.fn_c = callback
        elif callable(callback):
            self.fn_py = callback
        else:
            raise ValueError(f"Callback function {callback} wasn't of type OutputFuncWrapper or a callable object")


cdef class MaskedDetectorWorkspace(DetectorWorkspace):
    """Specialised workspace for detectors which support masking of modes.

    This workspace provides attributes that are exposed to both C and Python. The
    sections below detail how to use these for some workspace instance ``ws`` which
    inherits from ``MaskedDetectorWorkspace``.

    .. rubric:: Using via Python

    The ``unmasked_indices_arr`` attribute is a :class:`numpy.ndarray`, of dtype
    ``np.intp``, which contains the indices of modes which are not masked. One
    may then simply loop over this array of indices to access the corresponding field
    indices, e.g

    .. code-block:: python

        for k in ws.unmasked_indices_arr:
            # Do something with k, e.g. get field at 0 Hz freq. offset
            # at the given node for the mode index k:
            a_0k = carrier.get_out_fast(ws.dc_node_id, 0, k)
            # use a_0k for some calculation ...

    .. rubric:: Using via Cython

    This workspace also provides a ``unmasked_mode_indices`` pointer (only accessible
    from other Cython code) which corresponds to the data of the ``unmasked_indices_arr``
    NumPy array described above. The attribute ``num_unmasked_homs`` is the size of
    this array; i.e. the number of modes which are not masked.

    One may then write an optimised loop from ``[0, num_unmasked_homs)``, e.g

    .. code-block:: cython

        cdef Py_ssize_t i, k
        cdef complex_t a_0k
        for i in range(ws.num_unmasked_homs):
            k = ws.unmasked_mode_indices[i]
            # Do something with k, e.g. get field at 0 Hz freq. offset
            # at the given node for the mode index k:
            a_0k = carrier.get_out_fast(ws.dc_node_id, 0, k)
            # use a_0k for some calculation ...

    where each ``k`` is then the index of the mode at position ``i`` in the unmasked
    indices array.

    .. note::

        If the detector mask is empty (i.e. no modes are being masked) then
        ``unmasked_indices_arr`` (and, correspondingly, ``unmasked_mode_indices``)
        will simply be an array from ``[0, Nhoms)`` where ``Nhoms`` is the total
        number of modes in the simulation.
    """
    def __init__(self, owner, BaseSimulation sim, values=None, *, oinfo=None, **kwargs):
        super().__init__(owner, sim, values, oinfo=oinfo, **kwargs)

        self.unmasked_mode_indices = NULL
        self.has_mask = False

        self.setup_mask()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef int setup_mask(self) except -1:
        cdef:
            int[:, ::1] mask
            Py_ssize_t N_mask
            Py_ssize_t i
            Py_ssize_t count = 0

            int n, m

        # Check if owner has the right attribute, as some non-masked detectors also own masked
        # detector workspaces, and will otherwise throw an error here
        self.has_mask = getattr(self.owner, "has_mask", False)

        cdef np.ndarray[Py_ssize_t, ndim=1, mode="c"] unmasked
        if not self.has_mask: # no mask so just copy across all mode indices
            self.num_unmasked_HOMs = self.sim.model_settings.num_HOMs
            unmasked = np.arange(self.num_unmasked_HOMs, dtype=np.intp)
            self.unmasked_indices_arr = unmasked
            self.unmasked_mode_indices = &unmasked[0]
        else:
            mask = self.owner.mask
            N_mask = mask.shape[0]
            if N_mask >= self.sim.model_settings.num_HOMs:
                raise RuntimeError(
                    f"Error in detector {self.owner} (Output `{self.oinfo.name}`):\n"
                    f"    Length of mask array ({N_mask}) greater than or equal to "
                    f"number of modes ({self.sim.model_settings.num_HOMs})"
                )

            self.num_unmasked_HOMs = self.sim.model_settings.num_HOMs - N_mask
            unmasked = np.arange(self.num_unmasked_HOMs, dtype=np.intp)
            self.unmasked_indices_arr = unmasked
            self.unmasked_mode_indices = &unmasked[0]

            for i in range(self.sim.model_settings.num_HOMs):
                n = self.sim.model_settings.homs_view[i][0]
                m = self.sim.model_settings.homs_view[i][1]

                if not in_mask(n, m, mask):
                    if count >= self.num_unmasked_HOMs:
                        raise RuntimeError(
                            f"Error in detector {self.owner.name}:\n"
                            f"    One or more modes in the mask array do not exist in "
                            "the model."
                        )
                    self.unmasked_mode_indices[count] = i
                    count += 1

        return 0

    cdef bint hom_in_modes(self, Py_ssize_t hom_idx) noexcept:
        """Check whether a HOM index is in the non-masked modes of the detector."""
        if not self.has_mask:
            return True

        cdef Py_ssize_t i
        for i in range(self.num_unmasked_HOMs):
            if self.unmasked_mode_indices[i] == hom_idx:
                return True

        return False



import pytest

import finesse
from finesse.components import Mirror
from finesse.exceptions import ModelClassAttributeError, ModelMissingAttributeError
from finesse.utilities.collections import OrderedSet


@pytest.fixture
def model() -> finesse.Model:
    return finesse.Model()


def test_phase_level_change(model):
    # Issue 506
    model.phase_config(zero_k00=True, zero_tem00_gouy=False)
    assert model._settings.phase_config.zero_k00 is True
    assert model._settings.phase_config.zero_tem00_gouy is False
    model.phase_config(zero_k00=False, zero_tem00_gouy=True)
    assert model._settings.phase_config.zero_k00 is False
    assert model._settings.phase_config.zero_tem00_gouy is True
    model.phase_config(zero_k00=True, zero_tem00_gouy=True)
    assert model._settings.phase_config.zero_k00 is True
    assert model._settings.phase_config.zero_tem00_gouy is True


def test_get_string(model):
    model = finesse.Model()
    model.parse("m m1")
    model.parse("gauss g1 m1.p1.o w=1 Rc=1")
    model.beam_trace()

    assert model.m1 is model.get("m1")
    assert model.m1 is model.get_element("m1")

    assert model.m1.p1 is model.get("m1.p1")
    assert model.m1.p2 is model.get("m1.p2")

    assert model.m1.p2.i is model.get("m1.p2.i")
    assert model.m1.p2.o is model.get("m1.p2.o")

    assert model.m1.p2.i.qx is model.get("m1.p2.i.qx")
    assert model.m1.p2.i.qy is model.get("m1.p2.i.qy")
    assert model.m1.p2.o.qx is model.get("m1.p2.o.qx")
    assert model.m1.p2.o.qy is model.get("m1.p2.o.qy")

    assert model.m1.R is model.get("m1.R")
    assert model.m1.T is model.get("m1.T")


def test_get_other_model(model):
    model = finesse.Model()
    model.parse("m m1")

    model2 = finesse.Model()
    model2.parse("m m1")

    assert model.m1 is model.get(model2.m1)
    assert model.m1 is model.get_element(model2.m1)

    assert model.m1.p1 is model.get(model2.m1.p1)
    assert model.m1.p2 is model.get(model2.m1.p2)

    assert model.m1.p2.i is model.get(model2.m1.p2.i)
    assert model.m1.p2.o is model.get(model2.m1.p2.o)

    assert model.m1.R is model.get(model2.m1.R)
    assert model.m1.T is model.get(model2.m1.T)


def test_reduce_model_path(model):
    model.parse("l l1 P=5")
    assert model._reduce_model_path(["l1", "P"]) == 5


def test_reduce_model_path_namespaces(model):
    model.parse(
        """\
    m m1
    m m2
    link(m1, m2)"""
    )
    assert (
        model._reduce_model_path(["m1_p1__m2_p1"], check_namespaces=True)
        == model.spaces.m1_p1__m2_p1
    )


def test_reduce_model_path_namespaces_missing(model):
    model.parse(
        """\
    m m1
    m m2
    link(m1, m2)"""
    )
    with pytest.raises(ModelMissingAttributeError):
        model._reduce_model_path(["m1_p1__m2_p1"], check_namespaces=False)


def test_reduce_model_path_missing(model):
    model.parse("l l1 P=5")
    with pytest.raises(ModelMissingAttributeError):
        model._reduce_model_path(["l1", "__missing"])


def test_reduce_model_path_class_attr(model):
    model.parse("l l1 P=5")
    with pytest.raises(ModelClassAttributeError):
        model._reduce_model_path(["l1", "set_output_field"])


def test_get_namespace_paths_empty(model):
    assert model._get_namespace_paths() == OrderedSet()


def test_get_namespace_paths_with_elements(model):
    model.parse(
        """
    m m1
    m m2
    link(m1, m2)
    readout_dc r1
    readout_dc r2
    wire w1 r1.DC r2.DC
    noise n1 r1.DC 5.0
    """
    )
    assert model._get_namespace_paths() == {".spaces", ".wires", ".noises"}


def test_get_namespaces_empty(model):
    assert model._get_namespaces() == []


def test_get_namespaces_with_elements(model):
    model.parse(
        """
    m m1
    m m2
    link(m1, m2)
    readout_dc r1
    readout_dc r2
    wire w1 r1.DC r2.DC
    noise n1 r1.DC 5.0
    """
    )
    assert model._get_namespaces() == [model.noises, model.spaces, model.wires]


def test_get_elements_of_type(model):
    model.parse(
        """
    m m1
    m m2
    l l1
    """
    )
    assert model.get_elements_of_type("m") == (model.m1, model.m2)
    assert model.get_elements_of_type(Mirror) == (model.m1, model.m2)
    assert model.get_elements_of_type(Mirror, "Laser") == (model.m1, model.m2, model.l1)



import pytest
from finesse.script.exceptions import KatScriptError
from testutils.text import escape_full


@pytest.mark.parametrize(
    "script,error",
    (
        (
            "fake_element el1 a=1 1 b=2",
            "\nline 1: positional argument follows keyword argument\n"
            "-->1: fake_element el1 a=1 1 b=2\n"
            "                           ^",
        ),
        (
            "fake_element el1 1 2 a=1 3",
            "\nline 1: positional argument follows keyword argument\n"
            "-->1: fake_element el1 1 2 a=1 3\n"
            "                               ^",
        ),
    ),
)
def test_positional_arg_after_keyword_arg_invalid(parser, script, error):
    with pytest.raises(KatScriptError, match=escape_full(error)):
        parser.parse(script)


@pytest.mark.parametrize(
    "script,error",
    (
        (
            "fake_element el1 1 a=2 b=",
            "\nline 1: missing value\n"
            "-->1: fake_element el1 1 a=2 b=\n"
            "                              ^",
        ),
        (
            "fake_element el1 1 a= 2b=3",
            "\nline 1: missing value\n"
            "-->1: fake_element el1 1 a= 2b=3\n"
            "                          ^",
        ),
    ),
)
def test_hanging_equals_kwarg_invalid(parser, script, error):
    with pytest.raises(KatScriptError, match=escape_full(error)):
        parser.parse(script)



"""Squeezer-type optical components for producing squeezed light inputs."""
from finesse.parameter import float_parameter

from finesse.components.general import (
    Connector,
    FrequencyGenerator,
    NoiseGenerator,
    NoiseType,
)
from finesse.components.node import NodeType, NodeDirection


@float_parameter(
    "db",
    "Squeezing in amplitude dB (6dB is factor of 2 reduction in noise)",
    units="dB",
)
@float_parameter("f", "Frequency to squeeze at", units="Hz")
@float_parameter("angle", "Squeezing angle", units="degrees")
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Squeezer(Connector, FrequencyGenerator, NoiseGenerator):
    """Represents a squeezer producing a squeezed-light beam with a given squeezing in
    decibels and angle. The upper and conjugate of the lower sidebands can be excited in
    a signal analysis by injecting into the `.upper` and `.lower_conj` signal nodes.

    Parameters
    ----------
    name : str
        Name of the newly created squeezer.

    db : float
        Squeezing factor (in amplitude decibels). 6dB gives a factor of 2 reduction
        in noise.

    f : float or :class:`.Frequency`, optional
        Frequency-offset of the squeezer from the default (in Hz) or
        :class:`.Frequency` object. Defaults to 0 Hz offset.

    angle : float, optional
        Squeezing angle (in degrees). Defaults to zero.
    """

    def __init__(self, name, db, f=0, angle=0):
        Connector.__init__(self, name)
        FrequencyGenerator.__init__(self)
        NoiseGenerator.__init__(self)

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)
        # Two drives to excite an upper and lower sideband
        # individually used for generating squeezing info
        self._add_port("upper", NodeType.ELECTRICAL)
        self.upper._add_node("i", NodeDirection.INPUT)
        self._add_port("lower_conj", NodeType.ELECTRICAL)
        self.lower_conj._add_node("i", NodeDirection.INPUT)

        self.db = db
        self.f = f
        self.angle = angle

        self._register_noise_output("P1o", self.p1.o, NoiseType.QUANTUM)
        self._register_node_coupling("UPPER_P1o", self.upper.i, self.p1.o)
        self._register_node_coupling("LOWER_P1o", self.lower_conj.i, self.p1.o)

    @property
    def lower(self):
        from finesse.utilities.misc import deprecation_warning

        deprecation_warning(
            "Squeezer lower node has been renamed to lower_conj as it excites the"
            " conjugate of the lower sideband",
            "3.0.0",
        )
        return self.lower_conj

    def _source_frequencies(self):
        return [self.f.ref]

    def _couples_noise(self, ws, node, noise_type, frequency_in, frequency_out):
        return (frequency_in.index == frequency_out.index) or (
            (frequency_in.audio_carrier_index == frequency_out.audio_carrier_index)
            and (frequency_in.audio_carrier_index == ws.fsrc_car_idx)
        )

    def _get_workspace(self, sim):
        from finesse.components.modal.squeezer import (
            squeezer_fill_qnoise,
            squeezer_fill_signal,
            SqueezerWorkspace,
        )

        ws = SqueezerWorkspace(self, sim)
        ws.fsrc_car_idx = -1

        if sim.signal:
            ws.node_id = sim.signal.node_id(self.p1.o)
            ws.signal.add_fill_function(
                squeezer_fill_signal, True
            )  # TODO sort out refill flag here
            ws.signal.set_fill_noise_function(NoiseType.QUANTUM, squeezer_fill_qnoise)
            fsrc = self.__find_src_freq(sim.carrier)
            # Didn't find a Frequency bin for this squeezer in carrier simulation
            if fsrc is None:
                raise Exception(
                    f"Could not find a frequency bin in carrier sim at {self.f} for {self}"
                )
            ws.fsrc_car_idx = fsrc.index
            # Find the sideband frequencies
            sb = tuple(
                (
                    f
                    for f in sim.signal.optical_frequencies.frequencies
                    if f.audio_carrier_index == fsrc.index
                )
            )
            if len(sb) != 2:
                raise Exception(
                    f"Only something other than two audio sidebands {sb} for carrier "
                    f"{fsrc}"
                )
            ws.fcar_sig_sb_idx = (sb[0].index, sb[1].index)

        # if sim.is_modal: self._update_tem_gouy_phases(sim)
        return ws

    def _couples_frequency(self, ws, connection, frequency_in, frequency_out):
        # The only connections we have are signal inputs to optical output
        # And all the inputs should generate any output.
        return True

    def __find_src_freq(self, sim):
        # if it's tunable we want to look for the symbol that is just this
        # lasers frequency, as it will be changing
        for f in sim.optical_frequencies.frequencies:
            if not self.f.is_changing:
                # Don't match changing frequency bins if ours won't match
                if not f.symbol.is_changing and (
                    f.f == self.f.value  # match potential param refs
                    or f.f == float(self.f.value)  # match numeric values
                ):
                    # If nothing is changing then we can just match freq values
                    return f
            else:
                # If our frequency is changing then we have to have a frequency bin that
                # matches our symbol
                if f.symbol == self.f.ref:
                    return f  # Simple case
        return None



"""Fuzzing utilities for tests."""

import os

import numpy as np
from hypothesis import assume
from hypothesis.strategies import (
    characters,
    complex_numbers,
    composite,
    dictionaries,
    floats,
    from_regex,
    integers,
    lists,
    one_of,
    recursive,
    sampled_from,
    text,
)

from finesse import Model
from finesse.script.spec import KATSPEC
from finesse.symbols import OPERATORS

# Deadline setting for fuzzing; see #292.
DEADLINE = None if os.getenv("windows_ci") else 1000


# Min/max sizes for types that get used by numpy integer arrays.
_INT_DTYPE_INFO = np.iinfo(np.int32)
MIN_CINT = _INT_DTYPE_INFO.min
MAX_CINT = _INT_DTYPE_INFO.max


_MODEL = Model()

# Names that cannot be used for elements.
RESERVED_NAMES = {*dir(_MODEL), *KATSPEC.reserved_names}


# FIXME: don't hard-code these.
BINARY_OPERATIONS = (
    "__add__",
    "__sub__",
    "__mul__",
    "__truediv__",
    "__floordiv__",
    "__pow__",
)


# Valid normal fraction for reflectivity etc.
positive_normal_float = floats(min_value=0, max_value=1, allow_nan=False)

# Positive, finite floats.
positive_finite_floats = floats(min_value=0, allow_nan=False, allow_infinity=False)

# Finite integers for use with KLU (see #119).
c_integer = integers(min_value=MIN_CINT, max_value=MAX_CINT)
positive_c_integer = integers(min_value=0, max_value=MAX_CINT)


@composite
def numbers(draw, operations=False):
    """Any type of number."""
    cmplx_kwargs = dict(
        min_magnitude=0, max_magnitude=1, allow_nan=False, allow_infinity=False
    )

    if operations:
        # Wrap complex numbers in Function objects.
        cmplx_strategy = complex_number_operations(**cmplx_kwargs)
    else:
        cmplx_strategy = complex_numbers(**cmplx_kwargs)

    return draw(
        one_of(
            floats(min_value=-1, max_value=1, allow_nan=False, allow_infinity=False),
            integers(min_value=-1, max_value=1),
            cmplx_strategy,
        )
    )


@composite
def rtl_sets(draw):
    """R, T, L adding up to 1."""
    R = draw(positive_normal_float, label="R")
    T = draw(positive_normal_float, label="T")
    L = draw(positive_normal_float, label="L")

    # Don't allow R, T and L to be too small.
    if R != 0:
        assume(R > 1e-3)
    if T != 0:
        assume(T > 1e-3)
    if L != 0:
        assume(L > 1e-3)

    total = R + T + L
    assume(total > 0)
    R /= total
    T /= total
    L /= total
    return R, T, L


@composite
def rtl_sets_two_vals(draw):
    """R, T, L adding up to 1."""
    a = draw(positive_normal_float)
    b = draw(positive_normal_float)

    total = a + b
    assume(total > 0)

    if total > 1.0:
        a /= total
        b /= total

    return a, b


laser_powers = positive_c_integer


@composite
def complex_number_operations(draw, **kwargs):
    number = draw(complex_numbers(**kwargs))
    operation = OPERATORS["__sub__"] if number.imag < 0 else OPERATORS["__add__"]
    return operation(number.real, abs(number.imag))


@composite
def open_expressions(draw, rhs):
    lval = draw(numbers())
    operation = draw(sampled_from(BINARY_OPERATIONS))
    rval = draw(rhs)

    try:
        assume(operation not in ("/", "//") and rhs != 0)
    except TypeError:
        # Input type cannot be compared to zero.
        pass

    return OPERATORS[operation](lval, rval)


# TODO: make an Function for parenthesised expressions
# @composite
# def bracketed_expressions(draw, rhs):
#     return f"({draw(open_expressions(rhs))})"


@composite
def expressions(draw, rhs=None):
    rhs = numbers() if rhs is None else rhs
    # return draw(one_of(open_expressions(rhs), bracketed_expressions(rhs)))
    return draw(open_expressions(rhs))


@composite
def recursive_expressions(draw, max_leaves=25):
    """Generate nested expressions up to max_leaves."""
    return draw(
        recursive(
            expressions(), lambda rhs: expressions(rhs=rhs), max_leaves=max_leaves
        )
    )


@composite
def recursive_arrays(draw, operations=False):
    """Generate nested arrays up to max_leaves."""
    return [draw(recursive(numbers(operations=operations), lists))]


@composite
def line(draw, nonempty=False):
    """A single line of text."""
    # Blacklist control characters.
    return draw(
        text(
            min_size=1 if nonempty else 0,
            alphabet=characters(blacklist_categories=("Cs", "Cc")),
        )
    )


@composite
def kat_name(draw):
    """Valid kat script name."""
    name = draw(from_regex("^[a-zA-Z_][a-zA-Z0-9_]*$", fullmatch=True))
    assume(name not in RESERVED_NAMES)
    return name


@composite
def kat_empty(draw, multiline=False):
    if multiline:
        pattern = "^[ \t]*$"
    else:
        pattern = "^[ \t\n]*$"
    return draw(from_regex(pattern, fullmatch=True))


@composite
def kat_comment(draw):
    # Blacklist control characters.
    comment = draw(line())
    return f"#{comment}"


@composite
def kat_scalar(draw, nonnegative=False):
    """Kat script scalar value, including SI suffices."""
    min_value = 0 if nonnegative else None
    value = draw(floats(min_value=min_value, allow_nan=False, allow_infinity=False))
    items = [value]

    # Can't add a suffix for scientific notation.
    if "e" not in str(value):
        suffix = draw(sampled_from("pnumkMGT"))
        items.append(f"{value}{suffix}")

    return draw(sampled_from(items))


@composite
def kat_laser(draw):
    directive = draw(sampled_from(KATSPEC.elements["laser"].aliases))
    name = draw(kat_name())
    params = draw(
        dictionaries(keys=sampled_from(("P", "f", "phase")), values=kat_scalar())
    )
    paramstr = " ".join([f"{k}={v}" for k, v in params.items()])
    return f"{directive} {name} {paramstr}"


@composite
def kat_mirror(draw):
    directive = draw(sampled_from(KATSPEC.elements["mirror"].aliases))
    name = draw(kat_name())
    R, T, L = draw(rtl_sets())
    return f"{directive} {name} R={R} T={T} L={L}"


@composite
def kat_beamsplitter(draw):
    directive = draw(sampled_from(KATSPEC.elements["beamsplitter"].aliases))
    name = draw(kat_name())
    R, T, L = draw(rtl_sets())
    return f"{directive} {name} R={R} T={T} L={L}"


@composite
def kat_element(draw):
    return draw(one_of(kat_laser(), kat_mirror(), kat_beamsplitter()))


@composite
def kat_script_line(draw):
    """Generate kat script line."""
    return draw(one_of(kat_empty(), kat_comment(), kat_element()))



from finesse.cymath.cmatrix cimport SubCCSView2DArray
from finesse.simulations.workspace cimport ABCDWorkspace

from cpython.ref cimport PyObject

ctypedef object (*fptr_c_fill)(ConnectorWorkspace)

cpdef enum ConnectorCallbacks:
    FILL_MATRIX = 1,
    FILL_RHS = 2,
    FILL_NOISE = 4
    FILL_INPUT_NOISE = 8


cpdef enum ConnectionSetting:
    DISABLED = 0,
    DIAGONAL = 1,
    MATRIX = 2

cdef class FillFuncWrapper:
    cdef fptr_c_fill func
    @staticmethod
    cdef FillFuncWrapper make_from_ptr(fptr_c_fill f)

cdef struct node_noise_info:
    Py_ssize_t idx

cdef class NoiseInfo:
    cdef:
        int num_nodes
        SubCCSView2DArray nodes
        node_noise_info* node_info
        PyObject*** ptrs


ctypedef struct fill_info_t:
    # All callables should only take workspace as an argument
    PyObject *fn_py # Python callable
    fptr_c_fill fn_c # C callable
    bint refill


ctypedef struct fill_list_t:
    Py_ssize_t size
    Py_ssize_t num_refills # number of refills being done in this fill list
    fill_info_t[10] infos # 10 for now, can make this more adaptable with realloc if we want


cdef class ConnectorMatrixSimulationInfo:
    cdef:
        readonly object connections
        readonly object noise_sources

        fill_list_t matrix_fills

        FillFuncWrapper fn_rhs_c  # C RHS fill function
        object fn_rhs_py  # Python RHS fill function
        FillFuncWrapper fn_quantum_noise_c  # C noise input fill function
        object fn_quantum_noise_py  # Python noise input fill function
        FillFuncWrapper fn_quantum_noise_input_c  # C noise empty node input fill function
        object fn_quantum_noise_input_py  # Python noise empty node input fill function
        readonly ConnectorCallbacks callback_flag # flag stating which fill methods are to be called
        readonly dict connection_settings


cdef class ConnectorWorkspace(ABCDWorkspace):
    cdef readonly:
        ConnectorMatrixSimulationInfo carrier
        ConnectorMatrixSimulationInfo signal
        NoiseInfo input_noise
        NoiseInfo output_noise

    cdef setup_quantum_noise(self)



import finesse
import numpy as np


def test_issue_163():
    """power.pyx check_is_fsig wasn't working properly due to insufficient test cases
    between fsig/f symbols etc."""
    model = finesse.script.parse(
        """
    laser l1 P=1
    mod mod1 order=1 mod_type=pm f=80M midx=0.5
    m m1 R=0.5 L=0
    m m2 R=0.5 L=0
    link(l1,mod1,m1,10,m2)
    sgen sig m2.mech.z
    var dummy 1

    pd2 pd m1.p1.o 80M 0 100
    """
    )

    model.fsig.f = model.pd.f2.ref
    out0 = model.run("xaxis(pd.f2, log, 0.1, 100, 1)")

    # Check more complicated symbols
    model.fsig.f = -1 + model.pd.f2.ref + 1
    out1 = model.run("xaxis(pd.f2, log, 0.1, 100, 1)")

    model.fsig.f = 1 + model.dummy.ref - 1
    model.pd.f2 = model.dummy.ref
    out2 = model.run("xaxis(dummy, log, 0.1, 100, 1)")

    # Standard scanning
    model.fsig.f = 100
    model.pd.f2.value = model.fsig.f.ref
    out = model.run("xaxis(fsig, log, 0.1, 100, 1)")

    assert np.allclose(out["pd"], out0["pd"])
    assert np.allclose(out["pd"], out1["pd"])
    assert np.allclose(out["pd"], out2["pd"])


def test_pd0_sidebands():
    model = finesse.script.parse(
        """
    l l1 P=3
    mod mod1 f=10M midx=1 order=5
    link(l1, mod1)
    pd P mod1.p2.o
    """
    )
    sol = model.run()
    assert np.allclose(sol["P"], float(model.l1.P), atol=1e-14)



"""Collection of Actions that deal linear time invariant (LTI) modelling tasks."""

import warnings

from finesse.solutions import BaseSolution
from .base import Action, convert_str_to_parameter
import numpy as np
import logging
from finesse.simulations.sparse.simulation import SparseMatrixSimulation

# from finesse.components.readout import ReadoutDetectorOutput
from finesse.utilities.misc import is_iterable
from finesse.utilities import OrderedSet

LOGGER = logging.getLogger(__name__)


class OptimizationWarning(RuntimeWarning):
    pass


class OptimizeSolution(BaseSolution):
    """Solution for an optimization action.

    Attributes
    ----------
    result : scipy.optimize.optimize.OptimizeResult
        Result from the scipy optimization method that contains the results and some
        extra data about the process and any errors that might happen.

    parameters : [Parameter | tuple]
        Name or names of parameters that were optimized over

    x : [numeric | ndarray]
        The final minimized values for the parameters requested.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.result = None
        self.x = None
        self.parameters = None


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Optimize(Action):

    def __init__(
        self,
        detector,
        parameters,
        bounds=None,
        offset=0,
        kind="max",
        max_iterations=10000,
        tol=None,
        verbose=False,
        method="nelder-mead",
        opfunc=None,
        update_maps=False,
        pre_step=None,
        name="optimize",
        **kwargs,
    ):
        super().__init__(name)
        try:
            self.detector = detector.name
        except AttributeError:
            if is_iterable(detector):
                self.detector = tuple(str(_) for _ in detector)
            else:
                self.detector = str(detector)

        if isinstance(parameters, str):
            self.parameters = (parameters,)
        else:
            try:
                self.parameters = (*parameters,)
            except TypeError:
                self.parameters = (parameters,)

        self.bounds = bounds
        self.offset = offset
        self.kind = kind
        self.max_iterations = max_iterations
        self.tol = tol
        self.kwargs = kwargs
        self.verbose = verbose
        self.method = method
        self.update_maps = update_maps
        self.opfunc = opfunc
        self.pre_step = pre_step

    @property
    def parameter_names(self):
        return tuple(
            param if isinstance(param, str) else param.full_name
            for param in self.parameters
        )

    def _do(self, state):
        from scipy.optimize import minimize

        if state.sim is None:
            raise Exception("Simulation has not been built")
        if not isinstance(state.sim, SparseMatrixSimulation):
            raise NotImplementedError()

        out_wss = OrderedSet(  # workspaces can be in both lists
            (*state.sim.readout_workspaces, *state.sim.detector_workspaces)
        )
        if not isinstance(self.detector, str):
            dws = []
            for det in self.detector:
                for ws in out_wss:
                    if ws.oinfo.name == det:
                        dws.append(ws)
            if len(dws) != len(self.detector):
                raise RuntimeError(
                    f"Could not find a detector with the name {self.detector}"
                )
        else:
            dws = None
            for ws in out_wss:
                if ws.oinfo.name == self.detector:
                    dws = ws
                    break
            if dws is None:
                raise RuntimeError(
                    f"Could not find a detector with the name {self.detector}"
                )

        params = tuple(
            convert_str_to_parameter(state.model, param)
            for param in self.parameter_names
        )

        if len(params) == 0:
            raise RuntimeError(
                f"Could not find a parameter with the name {self.parameters}"
            )

        sol = OptimizeSolution(self.name)
        sol.iters = 0

        if self.opfunc is None:

            def func(x, params, state, dws):
                for a, p in zip(x, params):
                    p.value = a
                # Determine what the detector workspace needs to calculate
                # an output.
                if self.update_maps:
                    state.sim.update_map_data()
                if self.pre_step:
                    state.apply(self.pre_step)
                if dws.needs_carrier or dws.needs_signal or dws.needs_noise:
                    state.sim.run_carrier()
                if dws.needs_signal or dws.needs_noise:
                    state.sim.run_signal(dws.needs_noise)

                if self.kind == "max":
                    error = -np.abs(np.abs(dws.get_output()) + self.offset)
                else:
                    error = np.abs(np.abs(dws.get_output()) - self.offset)

                if self.verbose:
                    print(x, error)

                return error

            opfunc = func
        else:
            opfunc = self.opfunc

        res = minimize(
            opfunc,
            np.array([_.value for _ in params]),
            bounds=np.atleast_2d(self.bounds) if self.bounds else None,
            tol=self.tol,
            options={"maxiter": self.max_iterations, **self.kwargs},
            method=self.method,
            args=(params, state, dws),
        )
        if res.nit == 1:
            warnings.warn(
                "Optimisation finished after 1 iteration, tolerances might be too high",
                OptimizationWarning,
                stacklevel=1,
            )
        if self.verbose:
            if self.offset == 0:
                print(
                    f"Optimized {self.detector} to {res.fun:.6g} at {self.parameter_names} = {res.x[0]:.6g}"
                )
            else:
                print(
                    f"Optimized {self.detector} to {self.offset:g}{res.fun:+.6g} at {self.parameter_names} = {res.x[0]:.6g}"
                )
        # in case the values from the last iterations were not the optimal ones
        # explicitly set the parameter values to the optimal values
        for opt_value, param in zip(res.x, params):
            param.value = opt_value

        sol.result = res
        sol.x = res.x
        sol.parameters = self.parameter_names
        return sol

    def _requests(self, model, memo, first=True):
        memo["changing_parameters"].extend(self.parameter_names)

        # pass along requests needed from action called each step
        if self.pre_step:
            self.pre_step._requests(model, memo)

        # DDB temp comment out
        # if self.detector not in model.elements:
        #     raise RuntimeError(f"Could not find a detector called {self.detector}")

        # det = model.elements[self.detector]
        # if not (isinstance(det, ReadoutDetectorOutput)) and not (
        #     np.issubdtype(det.dtype, np.integer)
        #     or np.issubdtype(det.dtype, np.floating)
        #     or np.issubdtype(det.dtype, np.complexfloating)
        # ):
        #     raise RuntimeError(
        #         f"Detector {self.detector} must output a single integer or floating point value not {det.dtype}"
        #     )


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Minimize(Optimize):
    """An action that minimizes some detector output by applying some feedback
    to multiple targets in a model. Extra keyword arguments are passed on to the Scipy
    method:

    `minimize <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_

    This action offers a simplified interface that allows an optimization
    to be done during a simulation. By default the the Nelder-Mead optimization
    method is used but can be overridden. The user should read the Scipy documentation
    to determine which options should be used which are method dependant.

    Notes
    -----

    Default optimizer used is `nelder-mead`. To set the absolute and relative error
    targets use (From the scipy documentation:
    https://docs.scipy.org/doc/scipy/reference/optimize.minimize-neldermead.html)

    xatol : float, optional
        Absolute error in xopt between iterations that is acceptable for convergence.
        Defaults to 1e-4.

    fatol : float, optional
        Absolute error in func(xopt) between iterations that is acceptable for convergence.
        Defaults to 1e-4.

    These can be set as keyword arguments to the action.

    Parameters
    ----------

    detector : str
        The name of the detector output to maximize / minimize.

    parameter : [:class:`~.Parameter` | str | tuple]
        The parameter or name of the parameter to optimize, or a tuple of parameters
        when using multiple targets to optimize over.

    bounds : list, optional
        A pair of (lower, upper) bounds on the parameter value. Requires a method
        that uses bounds.

    offset : float, optional
        An offset applied to the detector output when optimizing, defaults to 0.

    kind : str, optional
        Either 'max' for maximization or 'min' for minimization, defaults to 'max'.

    max_iterations : int, optional
        Maximum number of solver iterations, defaults to 10000.

    method : str, optional
        Optimisation method to use, see Scipy documentation for options.

    name : str, optional
        The name of this action, defaults to 'maximize'.

    update_maps : bool, optional
        If you are changing some parameter or variable that a `Map` depends on
        then setting this flag to `True` will recompute the `Map` data for each
        iteration of the optimiser.

    pre_step : Action, optional
        Action to run on each step of the optimisation.

    kwargs
        Optional parameters passed to the Scipy optimisation routine as the
        `options` input. See Scipy method documentation to determine what is available.

    Examples
    --------

    Simple example that minimizes some measured power by feeding back to the laser
    power

    .. code-block:: python

        model = finesse.Model()
        model.parse('''
        l l1 P=1
        pd P l1.p1.o
        ''')
        sol = model.run("minimize(P, l1.P)")
        print(sol.result)
    """

    def __init__(self, detector, parameter, name="minimize", *args, **kwargs):
        super().__init__(detector, parameter, *args, name=name, kind="min", **kwargs)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Maximize(Optimize):
    """An action that maximizes some detector output by applying some feedback
    to multiple targets in a model. Extra keyword arguments are passed on to the Scipy
    method:

    `minimize <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_

    This action offers a simplified interface that allows an optimization
    to be done during a simulation. By default the the Nelder-Mead optimization
    method is used but can be overridden. The user should read the Scipy documentation
    to determine which options should be used which are method dependant.

    Notes
    -----

    Default optimizer used is `nelder-mead`. To set the absolute and relative error
    targets use (From the scipy documentation:
    https://docs.scipy.org/doc/scipy/reference/optimize.minimize-neldermead.html)

    xatol : float, optional
        Absolute error in xopt between iterations that is acceptable for convergence.
        Defaults to 1e-4.

    fatol : float, optional
        Absolute error in func(xopt) between iterations that is acceptable for convergence.
        Defaults to 1e-4.

    These can be set as keyword arguments to the action.

    Parameters
    ----------

    detector : str
        The name of the detector output to maximize / minimize.

    parameter : [:class:`~.Parameter` | str | tuple]
        The parameter or name of the parameter to optimize, or a tuple of parameters
        when using multiple targets to optimize over.

    bounds : list, optional
        A pair of (lower, upper) bounds on the parameter value. Requires a method
        that uses bounds.

    offset : float, optional
        An offset applied to the detector output when optimizing, defaults to 0.

    kind : str, optional
        Either 'max' for maximization or 'min' for minimization, defaults to 'max'.

    max_iterations : int, optional
        Maximum number of solver iterations, defaults to 10000.

    method : str, optional
        Optimisation method to use, see Scipy documentation for options.

    name : str, optional
        The name of this action, defaults to 'maximize'.

    update_maps : bool, optional
        If you are changing some parameter or variable that a `Map` depends on
        then setting this flag to `True` will recompute the `Map` data for each
        iteration of the optimiser.

    pre_step : Action, optional
        Action to run on each step of the optimisation.

    kwargs
        Optional parameters passed to the Scipy optimisation routine as the
        `options` input. See Scipy method documentation to determine what is available.

    Examples
    --------

    Simple example that maximizes the power in a coupled cavity solution by moving
    multiple mirrors

    .. code-block:: python

        model = finesse.Model()
        model.parse('''
        l l1 P=1
        m m1 R=0.98 T=0.02 phi=10
        m m2 R=0.99 T=0.01
        m m3 R=1 T=0 phi=-20
        link(l1, m1, m2, m3)
        pd P m3.p1.i
        ''')
        sol = model.run("maximize(P, [m1.phi, m3.phi], xatol=1e-7)")
        print(sol.result)
    """

    def __init__(self, detector, parameter, name="maximize", *args, **kwargs):
        super().__init__(detector, parameter, *args, name=name, kind="max", **kwargs)



"""This modules contains the functionality for building and running a sparse matrix
solver for computing the outputs of a simulation.

This is typically what earlier versions of FINESSE did in that two large sparse matrices
are built and solved for both the carrier and signal systems.
"""



"""Useful common utility functions and classes used throughout the Finesse package."""

# TODO: sort out which things get imported at the module level here
from .components import refractive_index
from .homs import make_modes, insert_modes
from .logging import logs, tracebacks
from .text import (
    ngettext,
    option_list,
    format_section,
    format_bullet_list,
    add_linenos,
    stringify,
    stringify_graph_gml,
)
from .misc import (
    check_name,
    pairwise,
    valid_name,
    is_iterable,
    opened_file,
    graph_layouts,
    networkx_layouts,
    graphviz_layouts,
)
from .units import SI, SI_LABEL, SI_VALUE
from .collections import OrderedSet
from .control import zpk_fresp
from finesse.utilities.numbers import clip_with_tolerance
from finesse.utilities.docs import class_to_url

__all__ = (
    "refractive_index",
    "make_modes",
    "insert_modes",
    "logs",
    "tracebacks",
    "ngettext",
    "option_list",
    "format_section",
    "format_bullet_list",
    "add_linenos",
    "stringify",
    "stringify_graph_gml",
    "check_name",
    "pairwise",
    "valid_name",
    "is_iterable",
    "opened_file",
    "graph_layouts",
    "networkx_layouts",
    "graphviz_layouts",
    "SI",
    "SI_LABEL",
    "SI_VALUE",
    "OrderedSet",
    "zpk_fresp",
    "clip_with_tolerance",
    "class_to_url",
)



#cython: boundscheck=False, wraparound=False, initializedcheck=False

cimport numpy as np
import numpy as np

from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport cabs, cexp, conj, creal, crotate
from finesse.cymath.math cimport isfinite, float_eq, nmax, radians, sqrt, NAN
from finesse.components.node import NodeType
from finesse.detectors.workspace cimport (
    DetectorWorkspace,
    OutputFuncWrapper,
)
from finesse.detectors.compute.power cimport (
    PD1Workspace,
    PD2Workspace,
    c_pd1_AC_output,
    c_pd2_AC_output,
)
from finesse.frequency cimport frequency_info_t, FrequencyContainer
from finesse.element_workspace cimport BaseCValues
from finesse.simulations.homsolver cimport HOMSolver
from finesse.simulations.sparse.solver cimport SparseSolver

cdef extern from "constants.h":
    long double H_PLANCK
    double ROOT2

ctypedef (double*, double*) ptr_tuple_2
ctypedef (double*, double*, double*, double*) ptr_tuple_4
ctypedef (double*, double*, double*, double*, double*, double*) ptr_tuple_6

# The names of these f, f1,f2,f3 etc. have to match the model parameters
# defined by the owner. Which is a bit annoying.

cdef class QND1Values(BaseCValues):
    cdef public:
        double f, phase

    def __init__(self):
        cdef ptr_tuple_2 ptr = (&self.f, &self.phase)
        cdef tuple params = ("f", "phase")
        self.setup(params, sizeof(ptr), <double**>&ptr)

cdef class QNDNValues(BaseCValues):
    cdef public:
        double f1, phase1
        double f2, phase2
        double f3, phase3

    def __init__(self):
        cdef ptr_tuple_6 ptr = (
                &self.f1, &self.phase1, &self.f2, &self.phase2, &self.f3, &self.phase3
                )
        cdef tuple params = ("f1", "phase1", "f2", "phase2", "f3", "phase3")
        self.setup(params, sizeof(ptr), <double**>&ptr)

cdef class QND0Workspace(DetectorWorkspace):
    cdef:
        complex_t[::1] s
        complex_t[::1] source_s
        complex_t[::1] cov
        PD1Workspace pd_ws
        int dc_node_id
        int ac_node_id
        int rhs_index
        int neq
        bint nsr

    def __init__(self, owner, sim, nsr, sources, exclude_sources):
        super().__init__(owner, sim, needs_noise=True)

        self.dc_node_id = sim.carrier.node_id(owner.node)
        self.ac_node_id = sim.signal.node_id(owner.node)
        self.neq = sim.signal.M().num_equations
        self.s = np.zeros(self.neq, dtype=np.complex128)
        self.source_s = np.empty(self.neq, dtype=np.complex128)
        self.set_output_fn(qnd0_output)

        self.rhs_index = self.owner._requested_selection_vectors[self.owner.name]

        self.nsr = nsr
        if self.nsr:
            self.pd_ws = PD1Workspace(owner, sim, sim.model.fsig.f, None)

        if sources:
            self.source_s[:] = 0
        elif exclude_sources:
            self.source_s[:] = 1
        if sources:
            self.fill_source_selection_vector(sources, 1)
        if exclude_sources:
            self.fill_source_selection_vector(exclude_sources, 0)

    cpdef fill_selection_vector(self) :
        cdef:
            HOMSolver carrier = self.sim.carrier
            HOMSolver signal = self.sim.signal
            Py_ssize_t i, j
            Py_ssize_t s_idx
            frequency_info_t fc, fs
            complex_t tmp

        self.update_parameter_values()
        for i in range(signal.optical_frequencies.size):
            fs = signal.optical_frequencies.frequency_info[i]
            fc = carrier.optical_frequencies.frequency_info[fs.audio_carrier_index]
            for j in range(nmax(signal.nhoms, 1)):
                s_idx = signal.field_fast(self.ac_node_id, fs.index, j)
                tmp = carrier.get_out_fast(self.dc_node_id, fc.index, j) * ROOT2
                if fs.audio_order > 0:
                    self.s[s_idx] = tmp
                else:
                    self.s[s_idx] = conj(tmp)

        for i in range(self.neq):
            signal.set_source_fast_3(i, self.s[i], self.rhs_index)

    cdef fill_source_selection_vector(self, sources, complex_t fill_value) :
        cdef:
            Py_ssize_t i, j
            Py_ssize_t node_id
            Py_ssize_t idx
            FrequencyContainer frequencies
            HOMSolver signal = self.sim.signal

        for comp in sources:
            for node_name in comp.nodes:
                if node_name not in signal.nodes:
                    continue
                node = signal.nodes[node_name]
                node_id = signal.node_id(node)
                if node.type == NodeType.OPTICAL:
                    for i in range(signal.optical_frequencies.size):
                        fs = signal.optical_frequencies.frequency_info[i]
                        for j in range(nmax(signal.nhoms, 1)):
                            idx = signal.field_fast(node_id, fs.index, j)
                            self.source_s[idx] = fill_value
                else:
                    if node.type == NodeType.MECHANICAL:
                        frequencies = signal.signal_frequencies[node]
                    elif node.type == NodeType.ELECTRICAL:
                        frequencies = signal.signal_frequencies[node]
                    else:
                        continue
                    for i in range(frequencies.size):
                        fs = frequencies.frequency_info[i]
                        idx = signal.field_fast(node_id, fs.index)
                        self.source_s[idx] = fill_value

    cpdef get_source_selection_vector(self) :
        return self.source_s

    cpdef set_covariance_matrix(self, complex_t[::1] v, unicode name) :
        # TODO: Should this be a copy?
        self.cov = v

qnd0_output = OutputFuncWrapper.make_from_ptr(c_qnd0_output)
cdef object c_qnd0_output(DetectorWorkspace self) :
    cdef:
        QND0Workspace ws = <QND0Workspace> self
        double rtn = 0
        double f0 = ws.sim.model_settings.f0
        double unit_vacuum = ws.sim.model_settings.UNIT_VACUUM
        double nf_factor = 0.25**1  # factor 0.25 per demodulation
        complex_t pdo
    for i in range(ws.neq):
        rtn += creal(ws.cov[i] * conj(ws.s[i]))
    # Compensate for demod 0.5 factor when demodulating a signal
    # frequency as this is what a network analyser would do.
    rtn *= 2

    if ws.nsr:
        pdo = c_pd1_AC_output(ws.pd_ws)
        rtn /= pdo.real * pdo.real + pdo.imag * pdo.imag

    return sqrt(
        unit_vacuum
        * rtn
        * H_PLANCK
        * f0
        * nf_factor
    )

cdef class QNDNWorkspace(DetectorWorkspace):
    cdef:
        complex_t[::1] s
        complex_t[::1] source_s
        complex_t[::1] cov
        DetectorWorkspace pd_ws
        int num_demod
        int Ndm
        int Nf
        int dc_node_id
        int ac_node_id
        int rhs_index
        long[:, ::1] demod_vac_contri
        long[:, ::1] demod_vac_contri_phi
        long[::1] demod_f_sig
        double[::1] demod_f
        double[::1] demod_phi
        QNDNValues cvalues
        bint nsr

        double[::1] freqs
        double[::1] phases
        int neq
        list carrier_demods

    def __init__(self, owner, sim, carrier_demods, nsr, sources, exclude_sources):
        assert(sim.signal is not None)
        assert(isinstance(sim.signal, SparseSolver))

        self.carrier_demods = carrier_demods
        num_carrier_demod = len(carrier_demods)
        if num_carrier_demod == 1:
            _values = QND1Values()
        else:
            _values = QNDNValues()
        super().__init__(owner, sim, _values, needs_noise=True)
        self.cvalues = <QNDNValues>self.values

        self.dc_node_id = self.sim.carrier.node_id(owner.node)
        self.ac_node_id = self.sim.signal.node_id(owner.node)
        self.set_output_fn(qndN_output)

        self.neq = self.sim.signal.M().num_equations
        # N demodulations, so we have a maximum of 2**N signal frequencies for each carrier
        # frequency, and each signal frequency can only have a maximum of 2**N contributing carrier
        # frequencies
        self.num_demod = num_carrier_demod + 1
        self.Ndm = int(2**self.num_demod)
        self.Nf = self.Ndm * self.sim.carrier.optical_frequencies.size
        self.demod_vac_contri = np.empty((self.Nf, self.Ndm), dtype=np.dtype("long"))
        self.demod_vac_contri_phi = np.empty((self.Nf, self.Ndm), dtype=np.dtype("long"))
        self.demod_f = np.full(self.Nf, np.nan, dtype=np.float64)
        self.demod_f_sig = np.zeros(self.Nf, dtype=np.dtype("long"))
        self.demod_phi = np.full(self.Ndm, np.nan, dtype=np.float64)
        self.s = np.zeros(self.neq, dtype=np.complex128)
        self.source_s = np.empty(self.neq, dtype=np.complex128)

        self.rhs_index = self.owner._requested_selection_vectors[self.owner.name]

        self.freqs = np.empty(self.num_demod, dtype=np.float64)
        self.phases = np.empty(self.num_demod, dtype=np.float64)

        self.nsr = nsr
        if self.nsr:
            # ddb todo: add more options here
            if num_carrier_demod == 1:
                # 1RF + 1fsig demodulation
                self.pd_ws = PD2Workspace(
                    owner,
                    sim,
                    carrier_demods[0][0],
                    carrier_demods[0][1],
                    sim.model.fsig.f,
                    None
                )
            else:
                raise NotImplementedError(f"Not yet implemented to calculate NSR for {num_carrier_demod} RF carrier demodulations")

        if sources:
            self.source_s[:] = 0
        elif exclude_sources:
            self.source_s[:] = 1
        if sources:
            self.fill_source_selection_vector(sources, 1)
        if exclude_sources:
            self.fill_source_selection_vector(exclude_sources, 0)

    cdef fill_carrier_qnoise_contributions(self) :
        cdef:
            Py_ssize_t i, j, k, nf
            Py_ssize_t base_idx
            frequency_info_t f
            double freq
            HOMSolver signal = self.sim.signal

        self.freqs[0] = self.cvalues.f1
        self.phases[0] = self.cvalues.phase1
        if self.num_demod > 2:
            self.freqs[1] = self.cvalues.f2
            self.phases[1] = self.cvalues.phase2
        if self.num_demod > 3:
            self.freqs[2] = self.cvalues.f3
            self.phases[2] = self.cvalues.phase3
        self.freqs[self.num_demod - 1] = self.sim.model_settings.fsig
        self.phases[self.num_demod - 1] = 0  # fsig demod doesn't affect result, so assume 0

        self.demod_phi[:] = 0
        for i in range(self.Ndm):
            for j in range(self.num_demod):
                if (i >> j) & 0x01:
                    self.demod_phi[i] += self.phases[j]
                else:
                    self.demod_phi[i] -= self.phases[j]
            self.demod_phi[i] = radians(self.demod_phi[i])

        self.demod_f[:] = 0
        for i in range(self.sim.carrier.optical_frequencies.size):
            f = self.sim.carrier.optical_frequencies.frequency_info[i]
            base_idx = i * self.Ndm
            for j in range(self.Ndm):
                self.demod_f[base_idx + j] = f.f
                for k in range(self.num_demod):
                    if (j >> k) & 0x01:
                        self.demod_f[base_idx + j] += self.freqs[k]
                    else:
                        self.demod_f[base_idx + j] -= self.freqs[k]

        # Use -1 to signal empty entries
        self.demod_f_sig[:] = -1
        self.demod_vac_contri[:] = -1

        # Check all frequencies to see if they match up to any signal sidebands
        for i in range(self.sim.carrier.optical_frequencies.size):
            f = self.sim.carrier.optical_frequencies.frequency_info[i]
            base_idx = i * self.Ndm
            for nf in range(self.Ndm):
                freq = self.demod_f[base_idx + nf]
                for j in range(self.Nf):
                    # Check frequencies against any existing signals
                    for k in range(signal.optical_frequencies.size):
                        if float_eq(self.demod_f[j], signal.optical_frequencies.frequency_info[k].f):
                            self.demod_f_sig[j] = k
                            break
                    if self.demod_f_sig[j] != -1:
                        continue
                    # If no signal frequency could be found then this is a pure
                    # vacuum noise field so need to add it to the contribution list
                    if float_eq(freq, self.demod_f[j]):
                        # If we have already found another item with this
                        # carrier frequency then these noises are correlated
                        for k in range(self.Ndm):
                            if self.demod_vac_contri[j][k] == -1:
                                self.demod_vac_contri[j][k] = f.index
                                self.demod_vac_contri_phi[j][k] = nf
                                break
                        break

    cpdef fill_selection_vector(self) :
        cdef:

            Py_ssize_t i, j, k
            Py_ssize_t f_idx, s_idx
            frequency_info_t fc, fs
            complex_t tmp
            HOMSolver signal = self.sim.signal
            HOMSolver carrier = self.sim.carrier

        self.update_parameter_values()
        self.fill_carrier_qnoise_contributions()

        self.s[:] = 0
        # For each carrier field check if the corresponding demodulated
        # frequency is a signal sideband. If so, we need to include it in the
        # selection vector.
        for i in range(self.sim.carrier.optical_frequencies.size):
            fc = self.sim.carrier.optical_frequencies.frequency_info[i]
            for j in range(self.Ndm):
                # If this is a signal sideband frequency this is the
                # product between the i'th carrier and this.
                f_idx = self.demod_f_sig[i * self.Ndm + j]
                if f_idx >= 0:
                    fs = signal.optical_frequencies.frequency_info[f_idx]
                    for k in range(nmax(signal.nhoms, 1)):
                        s_idx = signal.field_fast(self.ac_node_id, fs.index, k)
                        tmp = carrier.get_out_fast(self.dc_node_id, fc.index, k) * ROOT2
                        tmp = crotate(tmp, self.demod_phi[j])
                        if fs.audio_order > 0:
                            self.s[s_idx] = self.s[s_idx] + tmp
                        else:
                            self.s[s_idx] = self.s[s_idx] + conj(tmp)

        for i in range(self.neq):
            signal.set_source_fast_3(i, self.s[i], self.rhs_index)

    cdef fill_source_selection_vector(self, sources, complex_t fill_value) :
        cdef:
            Py_ssize_t i, j
            Py_ssize_t node_id
            Py_ssize_t idx
            FrequencyContainer frequencies
            HOMSolver signal = self.sim.signal

        for comp in sources:
            for node_name in comp.nodes:
                if node_name not in signal.nodes:
                    continue
                node = signal.nodes[node_name]
                node_id = signal.node_id(node)
                if node.type == NodeType.OPTICAL:
                    for i in range(signal.optical_frequencies.size):
                        fs = signal.optical_frequencies.frequency_info[i]
                        for j in range(nmax(signal.nhoms, 1)):
                            idx = signal.field_fast(node_id, fs.index, j)
                            self.source_s[idx] = fill_value
                else:
                    if node.type == NodeType.MECHANICAL:
                        frequencies = signal.signal_frequencies[node]
                    elif node.type == NodeType.ELECTRICAL:
                        frequencies = signal.signal_frequencies[node]
                    else:
                        continue
                    for i in range(frequencies.size):
                        fs = frequencies.frequency_info[i]
                        idx = signal.field_fast(node_id, fs.index, 0)
                        self.source_s[idx] = fill_value

    cpdef get_source_selection_vector(self) :
        return self.source_s

    cpdef set_covariance_matrix(self, complex_t[::1] v, unicode name) :
        # TODO: Should this be a copy?
        self.cov = v


qndN_output = OutputFuncWrapper.make_from_ptr(c_qndN_output)
cdef object c_qndN_output(DetectorWorkspace self) :
    cdef:
        QNDNWorkspace ws = <QNDNWorkspace> self
        HOMSolver carrier = ws.sim.carrier
        Py_ssize_t i, j, k
        Py_ssize_t Nhom = max(ws.sim.signal.nhoms, 1)
        double rtn = 0
        double f
        double f0 = ws.sim.model_settings.f0
        double unit_vacuum = ws.sim.model_settings.UNIT_VACUUM
        double nf_factor = 0.25**ws.num_demod  # factor 0.25 per demodulation
        complex_t pdo
        complex_t car_sum

    for i in range(ws.neq):
        rtn += creal(ws.cov[i] * conj(ws.s[i]))

    # Pick up pure vacuum noise and demodulate that into our signal.
    for i in range(ws.Nf):
        f = ws.demod_f[i]
        car_sum = 0
        for j in range(Nhom):
            for k in range(ws.Ndm):
                if ws.demod_vac_contri[i][k] == -1:
                    break
                car_sum += crotate(
                        carrier.get_out_fast(ws.dc_node_id, ws.demod_vac_contri[i][k], j),
                        # FIXME: There should be a minus sign here, why isn't there?
                        ws.demod_phi[ws.demod_vac_contri_phi[i][k]]
                        )
            rtn += cabs(car_sum)**2 * (1 + f / f0)

    if ws.nsr:
        if ws.num_demod == 2:
            (<PD2Workspace>ws.pd_ws).cvalues.f1 = ws.cvalues.f1
            (<PD2Workspace>ws.pd_ws).cvalues.phase1 = ws.cvalues.phase1
            (<PD2Workspace>ws.pd_ws).cvalues.f2 = 0
            (<PD2Workspace>ws.pd_ws).cvalues.phase2 = 0
        else:
            raise ValueError(f"Can't calculate NSR for {ws.num_demod-1} RF demodulations")
        pdo = c_pd2_AC_output(ws.pd_ws)
        rtn /= pdo.real * pdo.real + pdo.imag * pdo.imag

    # Compensate for demod 0.5 factor when demodulating a signal
    # frequency as this is what a network analyser would do.
    rtn *= 2

    return sqrt(
        unit_vacuum
        * rtn
        * H_PLANCK
        * f0
        * nf_factor
    )

cdef class QShot0Workspace(DetectorWorkspace):
    cdef:
        PD1Workspace pd_ws
        int Nf
        int dc_node_id
        int ac_node_id
        long[:, ::1] demod_vac_contri
        double[::1] demod_f
        bint nsr

    def __init__(self, owner, sim, nsr, *, output_info=None):
        assert(sim.signal is not None)
        super().__init__(owner, sim, oinfo=output_info, needs_noise=True)

        self.dc_node_id = self.sim.carrier.node_id(self.oinfo.nodes[0])
        self.ac_node_id = self.sim.signal.node_id(self.oinfo.nodes[0])
        self.set_output_fn(qshot0_output)

        # As we're only doing 1 demodulation, we have a maximum of 2 signal frequencies for each
        # carrier frequency, and each signal frequency can only have a maximum of 2 contributing
        # carrier frequencies
        self.Nf = 2 * self.sim.carrier.optical_frequencies.size
        self.demod_vac_contri = np.empty((self.Nf, 2), dtype=np.dtype("long"))
        self.demod_f = np.full(self.Nf, np.nan, dtype=np.float64)

        self.nsr = nsr
        if self.nsr:
            self.pd_ws = PD1Workspace(owner, sim, sim.model.fsig.f, None)

    cdef fill_carrier_qnoise_contributions(self) :
        cdef:
            Py_ssize_t i
            frequency_info_t f
            double freq
            double fsig = self.sim.model_settings.fsig

        # Use NAN / -1 to signal empty entries
        self.demod_f[:] = NAN
        self.demod_vac_contri[:] = -1

        for i in range(self.sim.carrier.optical_frequencies.size):
            f = self.sim.carrier.optical_frequencies.frequency_info[i]

            freq = f.f - fsig
            for j in range(self.Nf):
                if not isfinite(self.demod_f[j]):
                    self.demod_f[j] = freq
                    self.demod_vac_contri[j][0] = f.index
                    break
                elif float_eq(freq, self.demod_f[j]):
                    self.demod_vac_contri[j][1] = f.index
                    break

            freq = f.f + fsig
            for j in range(self.Nf):
                if not isfinite(self.demod_f[j]):
                    self.demod_f[j] = freq
                    self.demod_vac_contri[j][0] = f.index
                    break
                elif float_eq(freq, self.demod_f[j]):
                    self.demod_vac_contri[j][1] = f.index
                    break


qshot0_output = OutputFuncWrapper.make_from_ptr(c_qshot0_output)
cdef object c_qshot0_output(DetectorWorkspace self) :
    cdef:
        QShot0Workspace ws = <QShot0Workspace> self
        HOMSolver carrier = ws.sim.carrier
        Py_ssize_t i, j
        Py_ssize_t Nhom = max(ws.sim.signal.nhoms, 1)
        double rtn = 0
        double f
        double f0 = ws.sim.model_settings.f0
        double unit_vacuum = ws.sim.model_settings.UNIT_VACUUM
        double nf_factor = 0.25**1  # factor 0.25 per demodulation
        complex_t c1, c2
        complex_t pdo

    ws.fill_carrier_qnoise_contributions()

    # Pick up pure vacuum noise and demodulate that into our signal.
    for i in range(ws.Nf):
        f = ws.demod_f[i]
        if ws.demod_vac_contri[i][0] == -1:
            break
        if ws.demod_vac_contri[i][1] == -1:
            for j in range(Nhom):
                c1 = carrier.get_out_fast(ws.dc_node_id, ws.demod_vac_contri[i][0], j)
                rtn += cabs(c1)**2 * (1 + f / f0)
        else:
            for j in range(Nhom):
                c1 = carrier.get_out_fast(ws.dc_node_id, ws.demod_vac_contri[i][0], j)
                c2 = carrier.get_out_fast(ws.dc_node_id, ws.demod_vac_contri[i][1], j)
                rtn += cabs(c1 + c2)**2 * (1 + f / f0)

    # Compensate for demod 0.5 factor when demodulating a signal
    # frequency as this is what a network analyser would do.
    rtn *= 2

    if ws.nsr:
        pdo = c_pd1_AC_output(ws.pd_ws)
        rtn /= pdo.real * pdo.real + pdo.imag * pdo.imag

    return sqrt(
        unit_vacuum
        * rtn
        * H_PLANCK
        * f0
        * nf_factor
    )


cdef class QShotNWorkspace(DetectorWorkspace):
    """An N RF demodulation quantum shot noise detector workspace.

    Parameters
    ----------
    owner : object
        The model element that owns this workspace
    sim : object
        The current Simulation object that this workspace will use
        to generate its outputs
    carrier_demods : list
        A list of (frequency, phase) pairs. The frequency and phase
        object should be
    nsr : bool
        If True, the signal transfer function is computed and then
        used to compute the noise in equivalent units of the
        singal being injected at runtime.
    """
    cdef:
        DetectorWorkspace pd_ws
        list carrier_demods
        int num_demod
        int Ndm
        int Nf
        int dc_node_id
        int ac_node_id
        long[:, ::1] demod_vac_contri
        long[:, ::1] demod_vac_contri_phi
        double[::1] demod_f
        double[::1] demod_phi
        QNDNValues cvalues
        bint nsr

        double[::1] freqs
        double[::1] phases
        double[::1] tmp_fs


    def __init__(self, owner, sim, carrier_demods : list, nsr, *, output_info=None):
        assert(sim.signal is not None)
        self.carrier_demods = carrier_demods
        num_carrier_demod = len(carrier_demods)
        if num_carrier_demod == 1:
            _values = QND1Values()
        else:
            _values = QNDNValues()
        super().__init__(owner, sim, _values, oinfo=output_info, needs_noise=True)
        self.cvalues = <QNDNValues>self.values

        self.dc_node_id = self.sim.carrier.node_id(self.oinfo.nodes[0])
        self.ac_node_id = self.sim.signal.node_id(self.oinfo.nodes[0])
        self.set_output_fn(qshotN_output)

        # N demodulations, so we have a maximum of 2**N signal frequencies for each carrier
        # frequency, and each signal frequency can only have a maximum of 2**N contributing carrier
        # frequencies
        self.num_demod = num_carrier_demod + 1
        self.Ndm = int(2**self.num_demod)
        self.Nf = self.Ndm * self.sim.carrier.optical_frequencies.size
        self.demod_vac_contri = np.empty((self.Nf, self.Ndm), dtype=np.dtype("long"))
        self.demod_vac_contri_phi = np.empty((self.Nf, self.Ndm), dtype=np.dtype("long"))
        self.demod_f = np.full(self.Nf, np.nan, dtype=np.float64)
        self.demod_phi = np.full(self.Ndm, np.nan, dtype=np.float64)

        self.freqs = np.empty(self.num_demod, dtype=np.float64)
        self.phases = np.empty(self.num_demod, dtype=np.float64)
        self.tmp_fs = np.empty(self.Ndm, dtype=np.float64)

        self.nsr = nsr
        if self.nsr:
            # ddb todo: add more options here
            if num_carrier_demod == 1:
                # 1RF + 1fsig demodulation
                self.pd_ws = PD2Workspace(
                    owner,
                    sim,
                    carrier_demods[0][0],
                    carrier_demods[0][1],
                    sim.model.fsig.f,
                    None
                )
            else:
                raise NotImplementedError(f"Not yet implemented to calculate NSR for {num_carrier_demod} RF carrier demodulations")

    cdef fill_carrier_qnoise_contributions(self) :
        cdef:
            Py_ssize_t i, j, k
            frequency_info_t f
            double freq

        self.freqs[0] = self.cvalues.f1
        self.phases[0] = self.cvalues.phase1
        if self.num_demod > 2:
            self.freqs[1] = self.cvalues.f2
            self.phases[1] = self.cvalues.phase2
        if self.num_demod > 3:
            self.freqs[2] = self.cvalues.f3
            self.phases[2] = self.cvalues.phase3
        self.freqs[self.num_demod - 1] = self.sim.model_settings.fsig
        self.phases[self.num_demod - 1] = 0  # fsig demod doesn't affect result, so assume 0

        self.demod_phi[:] = 0
        for i in range(self.Ndm):
            for j in range(self.num_demod):
                if (i >> j) & 0x01:
                    self.demod_phi[i] += self.phases[j]
                else:
                    self.demod_phi[i] -= self.phases[j]
            self.demod_phi[i] = radians(self.demod_phi[i])

        # Use NAN / -1 to signal empty entries
        self.demod_f[:] = NAN
        self.demod_vac_contri[:] = -1

        for i in range(self.sim.carrier.optical_frequencies.size):
            f = self.sim.carrier.optical_frequencies.frequency_info[i]
            for j in range(self.Ndm):
                self.tmp_fs[j] = f.f
                for k in range(self.num_demod):
                    if (j >> k) & 0x01:
                        self.tmp_fs[j] += self.freqs[k]
                    else:
                        self.tmp_fs[j] -= self.freqs[k]

            for nf in range(self.Ndm):
                freq = self.tmp_fs[nf]
                for j in range(self.Nf):
                    if not isfinite(self.demod_f[j]):
                        self.demod_f[j] = freq
                        self.demod_vac_contri[j][0] = f.index
                        self.demod_vac_contri_phi[j][0] = nf
                        break
                    elif float_eq(freq, self.demod_f[j]):
                        for k in range(self.Ndm):
                            if self.demod_vac_contri[j][k] == -1:
                                self.demod_vac_contri[j][k] = f.index
                                self.demod_vac_contri_phi[j][k] = nf
                                break
                        break


qshotN_output = OutputFuncWrapper.make_from_ptr(c_qshotN_output)
cdef object c_qshotN_output(DetectorWorkspace self) :
    cdef:
        QShotNWorkspace ws = <QShotNWorkspace> self
        HOMSolver carrier = ws.sim.carrier
        Py_ssize_t i, j, k
        Py_ssize_t Nhom = max(ws.sim.signal.nhoms, 1)
        double rtn = 0
        double f
        double f0 = ws.sim.model_settings.f0
        double unit_vacuum = ws.sim.model_settings.UNIT_VACUUM
        double nf_factor = 0.25**ws.num_demod  # factor 0.25 per demodulation
        complex_t car_sum
        complex_t pdo

    ws.fill_carrier_qnoise_contributions()

    # Pick up pure vacuum noise and demodulate that into our signal.
    for i in range(ws.Nf):
        f = ws.demod_f[i]
        if ws.demod_vac_contri[i][0] == -1:
            break
        car_sum = 0
        for j in range(Nhom):
            for k in range(ws.Ndm):
                if ws.demod_vac_contri[i][k] == -1:
                    break
                car_sum += crotate(
                        carrier.get_out_fast(ws.dc_node_id, ws.demod_vac_contri[i][k], j),
                        # FIXME: There should be a minus sign here, why isn't there?
                        ws.demod_phi[ws.demod_vac_contri_phi[i][k]]
                        )
            rtn += cabs(car_sum)**2 * (1 + f / f0)

    # Compensate for demod 0.5 factor when demodulating a signal
    # frequency as this is what a network analyser would do.
    rtn *= 2

    if ws.nsr:
        if ws.num_demod == 2:
            (<PD2Workspace>ws.pd_ws).cvalues.f1 = ws.cvalues.f1
            (<PD2Workspace>ws.pd_ws).cvalues.phase1 = ws.cvalues.phase1
            (<PD2Workspace>ws.pd_ws).cvalues.f2 = 0
            (<PD2Workspace>ws.pd_ws).cvalues.phase2 = 0
        else:
            raise ValueError(f"Can't calculate NSR for {ws.num_demod-1} RF demodulations")
        pdo = c_pd2_AC_output(ws.pd_ws)
        rtn /= pdo.real * pdo.real + pdo.imag * pdo.imag

    return sqrt(
        unit_vacuum
        * rtn
        * H_PLANCK
        * f0
        * nf_factor
    )

cdef class QuantumNoiseDetectorWorkspace(DetectorWorkspace):
    cdef:
        long[:, ::1] demod_f_sig
        double[:, ::1] demod_f
        double[::1] demod_phi
        complex_t[::1] s
        complex_t[::1] v
        dict demod_vac_contri
        int dc_node_id
        int ac_node_id
        int Nf
        int Ndm
        int neq

    def __init__(self, owner, sim):
        assert(sim.signal is not None)
        super().__init__(owner, sim)
        for sim in sim:
            if sim.is_audio:
                self.sim.carrier = sim.sim.carrier
                self.sim.signal = sim

        self.dc_node_id = self.sim.carrier.node_id(owner.node)
        self.ac_node_id = self.sim.signal.node_id(owner.node)

        self.Nf = len(owner.freqs)
        self.Ndm = int(2 ** self.Nf)
        self.neq = self.sim.signal.M().num_equations

        self.demod_f = np.zeros((self.sim.carrier.optical_frequencies.size, self.Ndm))
        self.demod_f_sig = np.zeros((self.sim.carrier.optical_frequencies.size, self.Ndm), dtype=np.dtype("long"))
        self.demod_phi = np.zeros(self.Ndm)
        self.s = np.zeros(self.neq, dtype=np.complex128)

        self.set_output_fn(self.c_qnoised_output)

    cpdef fill_selection_vector(self) :
        cdef HOMSolver signal = self.sim.signal

        self.demod_f[:] = 0
        self.demod_f_sig[:] = 0
        self.demod_phi[:] = 0
        self.s[:] = 0

        self.fill_demod_f_table()
        self.fill_carrier_qnoise_contributions()
        self.fill_qnoised_selection_vector()

        for i in range(self.neq):
            signal.set_source_fast_3(i, self.s[i], self.owner.name)

    cpdef set_covariance_matrix(self, complex_t[::1] v, unicode name) :
        # TODO: Should this be a copy?
        self.v = v

    cdef void fill_demod_f_table(QuantumNoiseDetectorWorkspace self) noexcept:
        cdef:
            frequency_info_t f
            Py_ssize_t i, j, k

        for i in range(self.sim.carrier.optical_frequencies.size):
            f = self.sim.carrier.optical_frequencies.frequency_info[i]
            self.demod_f[f.index][:] = f.f

        for i in range(self.Ndm):
            for j in range(self.Nf):
                # Use bit shifting here to determine whether we add a plus
                # or minus to the total frequency and phases
                if (i >> j) & 0x01:
                    self.demod_phi[i] += self.owner.phases[j]
                    for k in range(self.sim.carrier.optical_frequencies.size):
                        f = self.sim.carrier.optical_frequencies.frequency_info[k]
                        self.demod_f[f.index][i] += float(self.owner.freqs[j])
                else:
                    self.demod_phi[i] -= self.owner.phases[j]
                    for k in range(self.sim.carrier.optical_frequencies.size):
                        f = self.sim.carrier.optical_frequencies.frequency_info[k]
                        self.demod_f[f.index][i] -= float(self.owner.freqs[j])

    cdef fill_carrier_qnoise_contributions(QuantumNoiseDetectorWorkspace self) :
        cdef:
            Py_ssize_t i, j, k
            Py_ssize_t car_idx
            frequency_info_t f
            HOMSolver signal = self.sim.signal

        self.demod_vac_contri = {}

        # Check all frequencies to see if they match up to any signal
        # sidebands
        for i in range(self.sim.carrier.optical_frequencies.size):
            car_idx = self.sim.carrier.optical_frequencies.frequency_info[i].index
            for j in range(self.Ndm):
                self.demod_f_sig[car_idx][j] = -1
                # Check frequencies against any existing signals, if we're not
                # only considering pure vacuum states
                if not self.owner.shot_only:
                    for k in range(signal.optical_frequencies.size):
                        f = signal.optical_frequencies.frequency_info[k]
                        if float_eq(self.demod_f[car_idx][j], f.f):
                            self.demod_f_sig[car_idx][j] = k
                            break

                # If no signal frequency could be found then this is a pure
                # vacuum noise field so need to add it to the contribution list
                if self.demod_f_sig[car_idx][j] == -1:
                    # Check if there are any other contributions listed already
                    # for this carrier
                    # TODO: remove direct comparison of floats
                    if self.demod_f[car_idx][j] in self.demod_vac_contri:
                        # If we have already found another item with this
                        # carrier frequency then these noises are correlated
                        elt = self.demod_vac_contri[self.demod_f[car_idx][j]]
                        elt["c_idx"].append(car_idx)
                        elt["phi_idx"].append(j)
                    else:
                        elt = {
                            "c_idx": [car_idx],
                            "phi_idx": [j],
                            "f": self.demod_f[car_idx][j],
                        }
                        self.demod_vac_contri[self.demod_f[car_idx][j]] = elt

    cdef fill_qnoised_selection_vector(QuantumNoiseDetectorWorkspace self) :
        cdef:
            Py_ssize_t i, j, k
            Py_ssize_t s_idx
            frequency_info_t fc, fs
            HOMSolver signal = self.sim.signal

        # For each carrier field check if the corresponding demodulated
        # frequency is a signal sideband. If so, we need to include it in the
        # selection vector.
        for j in range(self.sim.carrier.optical_frequencies.size):
            fc = self.sim.carrier.optical_frequencies.frequency_info[j]
            for k in range(self.Ndm):
                if self.demod_f_sig[j][k] >= 0:
                    # If this is a signal sideband frequency this is the
                    # product between the j'th carrier and this.
                    fs = signal.optical_frequencies.frequency_info[self.demod_f_sig[j][k]]
                    if fs.audio_order > 0:
                        for i in range(nmax(signal.nhoms, 1)):
                            s_idx = signal.field_fast(self.ac_node_id, fs.index, i)
                            self.s[s_idx] = self.s[s_idx] + (
                                self.sim.carrier.get_out_fast(self.dc_node_id, fc.index, i)
                                * ROOT2
                                * cexp(-1j * radians(self.demod_phi[k]))
                            )
                    else:
                        for i in range(nmax(signal.nhoms, 1)):
                            s_idx = signal.field_fast(self.ac_node_id, fs.index, i)
                            self.s[s_idx] = self.s[s_idx] + conj(
                                self.sim.carrier.get_out_fast(self.dc_node_id, fc.index, i)
                                * ROOT2
                                * cexp(-1j * radians(self.demod_phi[k]))
                            )

    cdef c_qnoised_output(self) :
        """Computes the output of the quantum noise detector.

        Returns
        -------
        np.complex128
            The output of this `QuantumNoiseDetector`.
        """
        cdef:
            double rtn = 0
            double f0 = self.sim.model_settings.f0
            double unit_vacuum = self.sim.model_settings.UNIT_VACUUM
            double nf_factor = 0.25**self.Nf
            HOMSolver signal = self.sim.signal
            HOMSolver carrier = self.sim.carrier

        if not self.owner.shot_only:
            for i in range(self.neq):
                rtn += creal(self.v[i] * conj(self.s[i]))

        # Now we must loop over the contributions from the demodulation for
        # frequencies that are not signal sidebands, i.e. we pick up pure
        # vacuum noise and demodulate that into our signal.
        for el in self.demod_vac_contri.values():
            if len(el["c_idx"]) > 1:
                for j in range(max(signal.nhoms, 1)):
                    val = 0
                    for k in el["c_idx"]:
                        val += (
                            carrier.get_out_fast(self.dc_node_id, k, j)
                        ).conjugate() * np.exp(
                            1j * np.deg2rad(self.demod_phi[<int>el["phi_idx"][j]])
                        )
                    if not self.owner.shot_only:
                        # TODO: why is this factor needed?
                        val *= np.sqrt(2)
                    rtn += abs(val)**2 * (1 + el["f"] / f0)
            else:
                val = 0
                for j in range(max(signal.nhoms, 1)):
                    val += abs(carrier.get_out_fast(self.dc_node_id, el["c_idx"][0], j)) ** 2
                rtn += val * (1 + el["f"] / f0)

        for f in self.owner.freqs:
            # Compensate for demod 0.5 factor if demodulating a signal
            # frequency as this is what a network analyser would do.
            if np.isclose(float(f), float(signal.model.fsig.f)):
                rtn *= 2

        return (
            unit_vacuum
            * rtn
            * H_PLANCK
            * f0
            * nf_factor
        ) ** 0.5





import logging

cimport numpy as np
import numpy as np
import cython
from libc.stdlib cimport free, calloc
from cpython.ref cimport PyObject
from finesse.cymath cimport complex_t
import finesse
from finesse.parameter cimport Parameter
from finesse.simulations.sparse.simulation cimport BaseSimulation
from finesse.simulations.homsolver cimport HOMSolver
from finesse.solutions.array import ArraySolution
from finesse.solutions.array cimport ArraySolution
from finesse.detectors.workspace import DetectorWorkspace
from finesse.detectors.workspace cimport DetectorWorkspace

from tqdm.auto import tqdm

LOGGER = logging.getLogger(__name__)


ctypedef object (*fptr_callback)(BaseSimulation)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef run_fsig_sweep(
        BaseSimulation sim,
        double[::1] axis,
        long[::1] input_node_indices,
        long[::1] output_node_indices,
        double[::1] input_scaling,
        double[::1] output_scaling,
        complex_t[:, :, ::1] out,
        bint compute_open_loop,
        tuple fsig_independant_outputs = None,
        tuple fsig_dependant_outputs = None,
    ):
    """Runs a simulation to sweep over a signal frequency axis. It does this in an
    optimised way for multiple inputs and outputs. It does not use detectors to
    compute outputs, it will just solve the matrix and return transfer functions
    between nodes. This is so it can be used internally for computing TFs without
    need to add detectors everywhere which the user has not specified.

    Parameters
    ----------
    sim : BaseSimulation
        The simulation object.
    axis : numpy.ndarray
        The signal frequency axis.
    input_node_indices : numpy.ndarray
        The indices of the input nodes.
    output_node_indices : numpy.ndarray
        The indices of the output nodes.
    input_scaling : numpy.ndarray
        The scaling factors for the input nodes.
    output_scaling : numpy.ndarray
        The scaling factors for the output nodes.
    out : numpy.ndarray
        The output array to store the transfer functions.
    compute_open_loop : bool
        A flag indicating whether to compute open loop transfer functions.
    fsig_independant_outputs : tuple, optional
        A tuple of fsig independent outputs.
    fsig_dependant_outputs : tuple, optional
        A tuple of fsig dependent outputs.

    Returns
    -------
    out : numpy.ndarray
        The transfer functions between nodes.
    other_outputs : dict, optional
        A dictionary of other outputs if fsig independent or dependent outputs are provided.
    """
    cdef:
        HOMSolver signal = sim.signal
        int Na = len(axis)
        int Ni = len(input_node_indices)
        int No = len(output_node_indices)
        int i, o, j
        complex_t denom
        Parameter f = sim.model.fsig.f
        bint cast_out = False
        DetectorWorkspace dws
        dict other_outputs = None

    for i in range(Ni):
        if not (0 <= input_node_indices[i] < signal.num_nodes):
            raise Exception(f"Input node index error: 0 <= {input_node_indices[i]} < {signal.num_nodes}")
    for o in range(No):
        if not (0 <= output_node_indices[o] < signal.num_nodes):
            raise Exception(f"Output node index error: 0 <= {output_node_indices[o]} < {signal.num_nodes}")

    if out is None:
        out = np.zeros((Na, No, Ni), dtype=np.complex128)
        cast_out = True
    else:
        assert(out.shape[0] == Na)
        assert(out.shape[1] == No)
        assert(out.shape[2] == Ni)

    if (fsig_independant_outputs is not None) or (fsig_dependant_outputs is not None):
        other_outputs = {}

    # We'll be making our own RHS inputs for this simulation
    signal.manual_rhs = True
    cdef double ifsig = sim.model_settings.fsig

    for j in range(Na):
        f.set_double_value(axis[j])
        sim.model_settings.fsig = axis[j]
        signal.refill()
        # For each output that is fsig independant get and store the output
        if fsig_independant_outputs:
            for dws in fsig_independant_outputs:
                other_outputs[dws.oinfo.name] = dws.get_output()

        for i in range(Ni):
            signal.clear_rhs()
            signal.set_source_fast(
                input_node_indices[i], 0, 0, input_scaling[i], 0
            )
            signal.solve()
            if not compute_open_loop:
                signal.get_out_fast
                for o in range(No):
                    out[j][o][i] = signal.get_out_fast(output_node_indices[o], 0, 0)

                    # scale output
                    out[j][o][i] = out[j][o][i] * output_scaling[o]
            else:
                for o in range(No):
                    out[j][o][i] = signal.get_out_fast(output_node_indices[o], 0, 0)

                    if input_node_indices[i] == output_node_indices[o]:
                        out[j][o][i] = out[j][o][i] - 1 # remove injected signal
                        out[j][o][i] = out[j][o][i]/(1+out[j][o][i])
                    else:
                        # We can divide out the 1/(1-H) closed loop behaviours by
                        # using the coupling computed back into the same input node
                        denom = signal.get_out_fast(input_node_indices[i], 0, 0) / input_scaling[i]
                        if denom.real == denom.imag == 0:
                            out[j][o][i] = 0
                        else:
                            out[j][o][i] = out[j][o][i] / denom

                    # scale output
                    out[j][o][i] = out[j][o][i] * output_scaling[o]

    signal.manual_rhs = False
    sim.model_settings.fsig = ifsig
    f.set_double_value(ifsig)

    if other_outputs is not None:
        if cast_out:
            return np.array(out), other_outputs
        else:
            return out, other_outputs
    else:
        if cast_out:
            return np.array(out)
        else:
            return out


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef run_fsig_sweep2(
        BaseSimulation sim,
        double[::1] axis,
        long[::1] input_node_indices,
        long[::1] input_freq_indices,
        long[::1] output_node_indices,
        double[::1] input_scaling,
        double[::1] output_scaling,
        complex_t[:, :, :, ::1] out,
        tuple fsig_independant_outputs = None,
        tuple fsig_dependant_outputs = None,
    ):
    """ Runs a simulation to sweep over a signal frequency axis.
    It does this in an optimised way for multiple inputs and outputs.

    `run_fsig_sweep2` differs to `run_fsig_sweep` in that the inputs should be
    optical nodes. The transfer functions from each HOM at the input to every
    output will then be calculated. Outputs should be some readout signal nodes.

    Transfer functions for lower audio sides must be requested to conjugate, as
    internally the conjugate of the lower is solved for.

    Returns
    -------
    transfer_functions : array_like
        shape of (frequencies, outputs, inputs, HOMs)
    """
    cdef:
        HOMSolver signal = sim.signal
        int Nm = signal.nhoms
        int Na = len(axis)
        int Ni = len(input_node_indices)
        int No = len(output_node_indices)
        int i, o, j, k
        Parameter f = sim.model.fsig.f
        bint cast_out = False
        DetectorWorkspace dws
        dict other_outputs = None

    if len(input_node_indices) != len(input_freq_indices):
        raise Exception("input node and frequency indices should be the same length")

    for i in range(Ni):
        if not (0 <= input_node_indices[i] < signal.num_nodes):
            raise Exception(f"Input node index error: 0 <= {input_node_indices[i]} < {signal.num_nodes}")
        if not (0 <= input_freq_indices[i] < signal.optical_frequencies.size):
            raise Exception(f"Input frequency index error: 0 <= {input_freq_indices[i]} < {signal.optical_frequencies.size}")

    for o in range(No):
        if not (0 <= output_node_indices[o] < signal.num_nodes):
            raise Exception(f"Output node index error: 0 <= {output_node_indices[o]} < {signal.num_nodes}")

    if out is None:
        out = np.zeros((Na, No, Ni, Nm), dtype=np.complex128)
        cast_out = True
    else:
        assert out.shape[0] == Ni
        assert out.shape[1] == Na
        assert out.shape[2] == No
        assert out.shape[3] == Nm

    if (fsig_independant_outputs is not None) or (fsig_dependant_outputs is not None):
        other_outputs = {}

    # We'll be making our own RHS inputs for this simulation
    signal.manual_rhs = True
    cdef double ifsig = sim.model_settings.fsig

    for j in range(Na):
        f.set_double_value(axis[j])
        sim.model_settings.fsig = axis[j]
        signal.refill()
        # For each output that is fsig independant get and store the output
        if fsig_independant_outputs:
            for dws in fsig_independant_outputs:
                other_outputs[dws.oinfo.name] = dws.get_output()

        for i in range(Ni):
            for k in range(Nm):
                signal.clear_rhs()
                # Loop over each mode at this node
                signal.set_source_fast(
                    input_node_indices[i], input_freq_indices[i], k, input_scaling[i], 0,
                )
                signal.solve()
                for o in range(No):
                    out[j][o][i][k] = signal.get_out_fast(output_node_indices[o], 0, 0)
                    # scale output
                    out[j][o][i][k] = out[j][o][i][k] * output_scaling[o]

    out *= np.sqrt(2) # normalise for the correct amplitude
    signal.manual_rhs = False
    sim.model_settings.fsig = ifsig
    f.set_double_value(ifsig)

    if other_outputs is not None:
        if cast_out:
            return np.array(out), other_outputs
        else:
            return out, other_outputs
    else:
        if cast_out:
            return np.array(out)
        else:
            return out


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef run_fsig_sweep3(
        BaseSimulation sim,
        double[::1] axis,
        long[::1] input_node_indices,
        long[::1] input_freq_indices,
        long[::1] output_node_indices,
        long[::1] output_freq_indices,
        double[::1] input_scaling,
        double[::1] output_scaling,
        complex_t[:, :, :, :, ::1] out,
        tuple fsig_independant_outputs = None,
        tuple fsig_dependant_outputs = None,
    ) :
    """ Runs a simulation to sweep over a signal frequency axis.
    It does this in an optimised way for multiple inputs and outputs.

    `run_fsig_sweep3` differs to `run_fsig_sweep` in that the input and output nodes
    should be optical nodes. The transfer functions from each HOM at the input to every
    output will then be calculated.

    Transfer functions for lower audio sides must be requested to conjugate, as
    internally the conjugate of the lower is solved for.

    Returns
    -------
    transfer_functions : array_like
        shape of (frequencies, outputs, inputs, HOMs, HOMs)
    """
    cdef:
        HOMSolver signal = sim.signal
        int Nm = signal.nhoms
        int Na = len(axis)
        int Ni = len(input_node_indices)
        int No = len(output_node_indices)
        int i, o, j, k, l
        Parameter f = sim.model.fsig.f
        bint cast_out = False
        DetectorWorkspace dws
        dict other_outputs = None

    if len(input_node_indices) != len(input_freq_indices):
        raise Exception("Input node and frequency indices should be the same length")
    if len(output_node_indices) != len(output_freq_indices):
        raise Exception("Output node and frequency indices should be the same length")

    for i in range(Ni):
        if not (0 <= input_node_indices[i] < signal.num_nodes):
            raise Exception(f"Input node index error: 0 <= {input_node_indices[i]} < {signal.num_nodes}")
        if not (0 <= input_freq_indices[i] < signal.optical_frequencies.size):
            raise Exception(f"Input frequency index error: 0 <= {input_freq_indices[i]} < {signal.optical_frequencies.size}")

    for o in range(No):
        if not (0 <= output_node_indices[o] < signal.num_nodes):
            raise Exception(f"Output node index error: 0 <= {output_node_indices[o]} < {signal.num_nodes}")
        if not (0 <= output_freq_indices[o] < signal.optical_frequencies.size):
            raise Exception(f"Output frequency index error: 0 <= {output_freq_indices[i]} < {signal.optical_frequencies.size}")

    if out is None:
        out = np.zeros((Na, No, Ni, Nm, Nm), dtype=np.complex128)
        cast_out = True
    else:
        assert(out.shape[0] == Na)
        assert(out.shape[1] == No)
        assert(out.shape[2] == Ni)
        assert(out.shape[3] == Nm) # output nodes
        assert(out.shape[4] == Nm) # input nodes

    if (fsig_independant_outputs is not None) or (fsig_dependant_outputs is not None):
        other_outputs = {}

    # We'll be making our own RHS inputs for this simulation
    signal.manual_rhs = True
    cdef double ifsig = sim.model_settings.fsig

    for j in range(Na):
        f.set_double_value(axis[j])
        sim.model_settings.fsig = axis[j]
        signal.refill()
        # For each output that is fsig independant get and store the output
        if fsig_independant_outputs:
            for dws in fsig_independant_outputs:
                other_outputs[dws.oinfo.name] = dws.get_output()

        for i in range(Ni):
            for k in range(Nm):
                signal.clear_rhs()
                # Loop over each mode at this node
                signal.set_source_fast(
                    input_node_indices[i], input_freq_indices[i], k, 1 * input_scaling[i], 0
                )
                signal.solve()
                for o in range(No):
                    for l in range(Nm):
                        # select output mode and scale output
                        out[j][o][i][l][k] = signal.get_out_fast(output_node_indices[o], output_freq_indices[o], l) * output_scaling[o]

    signal.manual_rhs = False
    sim.model_settings.fsig = ifsig
    f.set_double_value(ifsig)

    if other_outputs is not None:
        if cast_out:
            return np.array(out), other_outputs
        else:
            return out, other_outputs
    else:
        if cast_out:
            return np.array(out)
        else:
            return out


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef run_fsig_sweep4(
        BaseSimulation sim,
        double[::1] axis,
        long[::1] input_node_indices,
        long[::1] output_node_indices,
        long[::1] output_freq_indices,
        double[::1] input_scaling,
        double[::1] output_scaling,
        complex_t[:, :, :, ::1] out,
        tuple fsig_independant_outputs = None,
        tuple fsig_dependant_outputs = None,
    ):
    """ Runs a simulation to sweep over a signal frequency axis.
    It does this in an optimised way for multiple inputs and outputs.

    `run_fsig_sweep4` differs to `run_fsig_sweep` in that the input nodes are signal nodes
    and output nodes should be optical nodes. The transfer functions from each HOM at the
    input to every output will then be calculated.

    Transfer functions for lower audio sides must be requested to conjugate, as
    internally the conjugate of the lower is solved for.

    Returns
    -------
    transfer_functions : array_like
        shape of (frequencies, outputs, inputs, HOMs)
    """
    cdef:
        HOMSolver signal = sim.signal
        int Nm = signal.nhoms
        int Na = len(axis)
        int Ni = len(input_node_indices)
        int No = len(output_node_indices)
        int i, o, j, k
        Parameter f = sim.model.fsig.f
        bint cast_out = False
        DetectorWorkspace dws
        dict other_outputs = None

    if len(output_node_indices) != len(output_freq_indices):
        raise Exception("Output node and frequency indices should be the same length")

    for o in range(No):
        if not (0 <= output_node_indices[o] < signal.num_nodes):
            raise Exception(f"Output node index error: 0 <= {output_node_indices[o]} < {signal.num_nodes}")
        if not (0 <= output_freq_indices[o] < signal.optical_frequencies.size):
            raise Exception(f"Output frequency index error: 0 <= {output_freq_indices[o]} < {signal.optical_frequencies.size}")

    if out is None:
        out = np.zeros((Na, No, Ni, Nm), dtype=np.complex128)
        cast_out = True
    else:
        assert(out.shape[0] == Na)
        assert(out.shape[1] == No)
        assert(out.shape[2] == Ni)
        assert(out.shape[3] == Nm)

    if (fsig_independant_outputs is not None) or (fsig_dependant_outputs is not None):
        other_outputs = {}

    # We'll be making our own RHS inputs for this simulation
    signal.manual_rhs = True
    cdef double ifsig = sim.model_settings.fsig

    for j in range(Na):
        f.set_double_value(axis[j])
        sim.model_settings.fsig = axis[j]
        signal.refill()
        # For each output that is fsig independant get and store the output
        if fsig_independant_outputs:
            for dws in fsig_independant_outputs:
                other_outputs[dws.oinfo.name] = dws.get_output()

        for i in range(Ni):
            signal.clear_rhs()
            signal.set_source_fast(
                input_node_indices[i], 0, 0, input_scaling[i], 0
            )
            signal.solve()
            for o in range(No):
                # Loop over each mode at this node
                for k in range(Nm):
                    # select output mode and scale output
                    out[j][o][i][k] = signal.get_out_fast(output_node_indices[o], output_freq_indices[o], k) * output_scaling[o]

    out /= np.sqrt(2) # normalise for the correct amplitude
    signal.manual_rhs = False
    sim.model_settings.fsig = ifsig
    f.set_double_value(ifsig)

    if other_outputs is not None:
        if cast_out:
            return np.array(out), other_outputs
        else:
            return out, other_outputs
    else:
        if cast_out:
            return np.array(out)
        else:
            return out


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def run_axes_scan(
        object state,
        tuple axes,
        tuple params,
        double[:] offsets,
        tuple out_shape,
        ArraySolution solution,
        object pre_step,
        object post_step,
        bint progress_bar = False,
        str progress_bar_desc = ""
    ):
    cdef:
        BaseSimulation sim = state.sim
        int Np = len(params)
        int No = len(offsets)
        int Na = len(axes)
        int Nos = len(out_shape)
        int i, step
        np.ndarray[double, ndim=1, mode="c"] narr
        double** ptr_axes = NULL
        int* ptr_axes_len = NULL
        PyObject** ptr_params = NULL
        bint mask_this_point = False
        int[::1] IDX = np.zeros(Nos, dtype=np.int32)
        int Ntot

        Parameter param

    if(not (Np == No == Na == Nos)):
        raise Exception("Param, offsets, axes, and out_shape lengths are not the same")

    for p in params:
        if p.datatype not in (float, np.float64):
            raise Exception(f"Can only vary parameters with datatype float, not {p.full_name} with {p.datatype}")

    try:
        # Can't have a memory view of typed ndarray apparently.
        # So here we check the axes are double c-continuous and
        # then save the double pointer
        ptr_axes = <double**> calloc(Na, sizeof(double*))
        if not ptr_axes:
            raise MemoryError()

        ptr_axes_len = <int*> calloc(Na, sizeof(int))
        if not ptr_axes_len:
            raise MemoryError()

        for i in range(Na):
            narr = <np.ndarray[double, ndim=1, mode="c"]?> axes[i]
            if narr.size != out_shape[i]:
                raise Exception(f"Out shape[{i}]={out_shape[i]} is not the correct size for the axes[i]={narr.size}")

            ptr_axes[i] = &narr[0]
            ptr_axes_len[i] = narr.size

        # Then to get around some annoying python referencing and issues
        # with accessing cdefs of extension types in a memory view we
        # make an array of PyObjects
        ptr_params = <PyObject**> calloc(Np, sizeof(PyObject*))
        if not ptr_params:
            raise MemoryError()
        for i in range(Np):
            ptr_params[i] = <PyObject*>(<Parameter?>params[i])

        Ntot = np.prod(out_shape)

        if progress_bar:
            pbar = tqdm(range(Ntot), leave=False, desc=progress_bar_desc, disable=not finesse.config.show_progress_bars)
        else:
            pbar = None

        # Now iterate over the all the axes
        #for step in range(Ntot):
        #for idx in np.ndindex(*out_shape):
        for step in range(Ntot):
            for i in range(Np):
                (<Parameter>ptr_params[i]).set_double_value(ptr_axes[i][IDX[i]] + offsets[i])
            # ------------------------------------------------------
            # PRE STEP
            # ------------------------------------------------------
            if pre_step is not None:
                pre_step._do(state)
            # ------------------------------------------------------
            # DO STEP
            # ------------------------------------------------------
            mask_this_point = not sim.run_carrier()

            if not mask_this_point and sim.signal is not None:
                sim.run_signal()

            if progress_bar:
                pbar.update()
            # ------------------------------------------------------
            # POST STEP
            # ------------------------------------------------------
            if mask_this_point:
                values_str = ""
                for i in range(Np):
                    param = <Parameter>ptr_params[i]
                    if param.__units is not None:
                        param_units = " " + param.__units
                    else:
                        param_units = ""

                    values_str += (
                        param.__full_name
                        + " = "
                        + str(ptr_axes[i][IDX[i]] + offsets[i])
                        + param_units
                    )
                    if i != Np - 1:
                        values_str += ", "

                LOGGER.error("Masking simulation outputs at: %s", values_str)

            if solution.update(step, mask_this_point) == -1:
                raise RuntimeError("Exception calling solution update")

            if post_step is not None:
                post_step._do(state)

            # ------------------------------------------------------

            # Increment the index vector
            for i in range(No):
                i = Nos-i-1
                IDX[i] += 1
                if IDX[i] >= ptr_axes_len[i]:
                    IDX[i] = 0
                else:
                    break

    finally:
        # This forces pbar to show, even when leave=False
        if progress_bar:
            pbar.refresh()
            pbar.close()
        if ptr_axes != NULL: free(ptr_axes)
        if ptr_axes_len != NULL: free(ptr_axes_len)
        if ptr_params != NULL: free(ptr_params)



# import pytest
# from pathlib import Path

# import setup


# def test_check_pyx_files(tmpdir: Path, monkeypatch: pytest.MonkeyPatch):
#     tmpdir = Path(tmpdir)
#     monkeypatch.setattr(setup, "FINESSE_DIR", tmpdir)
#     fn = "foo.pyx"
#     pyx_file = Path(tmpdir / fn)
#     pyx_file.touch()
#     setup.check_pyx_files(pyx_files=[fn])


# def test_check_pyx_files_missing(tmpdir: Path, monkeypatch: pytest.MonkeyPatch):
#     tmpdir = Path(tmpdir)
#     monkeypatch.setattr(setup, "FINESSE_DIR", tmpdir)
#     fn = "foo.pyx"
#     with pytest.raises(FileNotFoundError):
#         setup.check_pyx_files(pyx_files=[fn])


# def test_check_pyx_files_ignore_hidden(tmpdir: Path, monkeypatch: pytest.MonkeyPatch):
#     tmpdir = Path(tmpdir)
#     monkeypatch.setattr(setup, "FINESSE_DIR", tmpdir)
#     fn = "foo.pyx"
#     pyx_file = tmpdir / fn
#     pyx_file.touch()
#     hidden_dir = tmpdir / ".ipynb_checkpoints"
#     hidden_dir.mkdir()
#     (hidden_dir / "bar.pyx").touch()
#     setup.check_pyx_files(pyx_files=[fn])
